#    "Commons Clause" License Condition v1.0
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

import json
import os
import platform
import re
import shutil
import sys
import tempfile
import urllib.request
import zipfile
import logging
from pathlib import Path
import revedaEditor.backend.editFunctions as edf

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QTableWidget,
    QTableWidgetItem,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QProgressBar,
    QTextEdit,
    QVBoxLayout,
    QWidget,
    QHeaderView,
    QApplication,
    QFileDialog,
)


class PDKRegistryWindow(QMainWindow):
    DEFAULT_REGISTRY = (
        "https://raw.githubusercontent.com/eskiyerli/revolutionEDA_pdks/main/pdks.json"
    )

    def __init__(self, parent=None, registry_url: str | None = None):
        super().__init__(parent)
        self.setWindowTitle("Revolution EDA PDK Registry")
        self.resize(900, 400)
        self.logger = logging.getLogger("reveda")

        self.registry_url = registry_url or self.DEFAULT_REGISTRY
        app = QApplication.instance()
        self.pdksDir = app.basePath.parent/'pdks'
        self.pdksDir = self.pdksDir.resolve()
        if not self.pdksDir.exists():
            self.pdksDir.mkdir(parents=True)

        self._initUI()
        self._registry = []
        self.fetch_registry()

    def _initUI(self):
        central = QWidget()
        self.setCentralWidget(central)
        main_l = QVBoxLayout(central)
        
        # PDK directory path display
        pathLayout = QHBoxLayout()
        pathLayout.addWidget(
            edf.boldLabel("Local PDKs Directory"), 1)
        self.pdkPathEdit = edf.longLineEdit()
        self.pdkPathEdit.setText(str(self.pdksDir))
        pathLayout.addWidget(self.pdkPathEdit, 5)
        self.pdkPathButton = QPushButton("...")
        self.pdkPathButton.clicked.connect(self.onPDKPathButtonClicked)
        pathLayout.addWidget(self.pdkPathButton, 1)
        main_l.addLayout(pathLayout)
        
        # Main content area
        content_l = QHBoxLayout()

        left_w = QWidget()
        left_l = QVBoxLayout(left_w)
        left_l.setContentsMargins(0, 0, 0, 0)
        self.tableWidget = QTableWidget()
        self.tableWidget.setColumnCount(6)
        self.tableWidget.setHorizontalHeaderLabels(
            ["Installed", "Binary","PDK", "Process", "Version", "License"]
        )
        self.tableWidget.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeMode.Stretch
        )
        self.tableWidget.setSelectionBehavior(QTableWidget.SelectionBehavior.SelectRows)
        self.tableWidget.setSelectionMode(QTableWidget.SelectionMode.SingleSelection)
        left_l.addWidget(QLabel("Available PDKs:"))
        left_l.addWidget(self.tableWidget)
        self.refresh_btn = QPushButton("Refresh")
        left_l.addWidget(self.refresh_btn)
        content_l.addWidget(left_w, 2)

        right_w = QWidget()
        right_l = QVBoxLayout(right_w)
        right_l.setContentsMargins(0, 0, 0, 0)
        right_l.addWidget(QLabel("Description:"))
        self.desc = QTextEdit()
        self.desc.setReadOnly(True)
        right_l.addWidget(self.desc, 1)
        self.download_btn = QPushButton("Download / Install")
        self.uninstall_btn = QPushButton("Uninstall")
        self.progress = QProgressBar()
        self.progress.setRange(0, 100)
        right_l.addWidget(self.progress)
        right_l.addWidget(self.download_btn)
        right_l.addWidget(self.uninstall_btn)
        content_l.addWidget(right_w, 2)
        
        main_l.addLayout(content_l)

        self.tableWidget.itemSelectionChanged.connect(self._onSelect)
        self.tableWidget.itemDoubleClicked.connect(self._on_item_activated)
        self.tableWidget.itemChanged.connect(self._on_checkbox_changed)
        self.download_btn.clicked.connect(self._on_download)
        self.uninstall_btn.clicked.connect(self._on_uninstall)
        self.refresh_btn.clicked.connect(self.fetch_registry)


    def onPDKPathButtonClicked(self):
        self.pdkPathEdit.setText(
            QFileDialog.getExistingDirectory(self, caption="Select PDK Repo Path")
        )
        self.pdksDir = Path(self.pdkPathEdit.text())

    def fetch_registry(self):
        self.tableWidget.setRowCount(0)
        self.desc.clear()
        self.progress.setValue(0)
        try:
            with urllib.request.urlopen(self.registry_url) as resp:
                raw = resp.read()
                index = json.loads(raw.decode("utf-8"))
                if isinstance(index, dict) and "pdks" in index:
                    index = index["pdks"]
                self._registry = index if isinstance(index, list) else []
        except Exception as exc:
            QMessageBox.warning(
                self, "Registry Error", f"Failed to fetch registry:\n{exc}"
            )
            self._registry = []

        self.tableWidget.setRowCount(len(self._registry))
        for row, entry in enumerate(self._registry):
            name = entry.get("name", "unknown")
            pdk_dir = self.pdksDir / re.sub(r"[^A-Za-z0-9_.-]", "_", name)

            # Check if installed and get local version
            config_path = pdk_dir / 'config.json'
            isInstalled = config_path.exists()
            isBinary = entry.get("type") == "binary"
            local_version = None
            
            if isInstalled:
                try:
                    with open(config_path) as f:
                        config = json.load(f)
                        local_version = config.get("pdk_version", "0.0.0")
                except Exception:
                    local_version = "0.0.0"
            
            # Compare versions
            remote_version = entry.get("pdk_version", "0.0.0")
            is_upgradeable = isInstalled and local_version and self._compare_versions(local_version, remote_version) < 0
            installedItem = QTableWidgetItem()
            installedItem.setFlags(Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled)
            installedItem.setCheckState(
                Qt.CheckState.Checked if isInstalled else Qt.CheckState.Unchecked
            )
            installedItem.setData(Qt.ItemDataRole.UserRole, entry)
            if is_upgradeable:
                installedItem.setText("⬆")
                installedItem.setToolTip(f"Update available: {local_version} → {remote_version}")

            binaryItem = QTableWidgetItem()
            binaryItem.setFlags(Qt.ItemFlag.ItemIsEnabled)
            binaryItem.setCheckState(
                Qt.CheckState.Checked if isBinary else Qt.CheckState.Unchecked
            )
            binaryItem.setData(Qt.ItemDataRole.UserRole, entry)
            self.tableWidget.setItem(row, 0, installedItem)
            self.tableWidget.setItem(row, 1, binaryItem)

            self.tableWidget.setItem(row, 2, QTableWidgetItem(name))
            self.tableWidget.setItem(
                row, 3, QTableWidgetItem(entry.get("process", "N/A"))
            )
            self.tableWidget.setItem(
                row, 4, QTableWidgetItem(entry.get("version", "0.0.0"))
            )
            self.tableWidget.setItem(
                row, 5, QTableWidgetItem(entry.get("license", "Unknown"))
            )

    def _compare_versions(self, v1: str, v2: str) -> int:
        """Compare two version strings. Returns -1 if v1 < v2, 0 if equal, 1 if v1 > v2"""
        try:
            parts1 = [int(x) for x in v1.split(".")]
            parts2 = [int(x) for x in v2.split(".")]
            for p1, p2 in zip(parts1, parts2):
                if p1 < p2:
                    return -1
                elif p1 > p2:
                    return 1
            
            if len(parts1) < len(parts2):
                return -1
            elif len(parts1) > len(parts2):
                return 1
            return 0
        except Exception:
            return 0

    def _on_checkbox_changed(self, item: QTableWidgetItem):
        if item.column() != 0:
            return
        entry = item.data(Qt.ItemDataRole.UserRole)
        if not entry:
            return
        
        name = re.sub(r"[^A-Za-z0-9_.-]", "_", entry.get("name", "pdk"))
        pdk_dir = self.pdksDir / name
        
        if item.checkState() == Qt.CheckState.Checked and not pdk_dir.exists():
            self._install_entry(entry)
        elif item.checkState() == Qt.CheckState.Unchecked and pdk_dir.exists():
            self._uninstall_pdk(name, pdk_dir)

    def _uninstall_pdk(self, name: str, pdk_dir: Path):
        try:
            # Remove main PDK directory
            shutil.rmtree(pdk_dir)
            
            # Remove any other files/directories containing PDK name
            for item in self.pdksDir.iterdir():
                if name.lower() in item.name.lower():
                    if item.is_dir():
                        shutil.rmtree(item)
                    else:
                        item.unlink()
            
            self.logger.info(f"PDK '{name}' and related files uninstalled.")
            self.fetch_registry()
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e))
            self.fetch_registry()

    def _onSelect(self):
        self.desc.clear()
        current_row = self.tableWidget.currentRow()
        if current_row < 0:
            return
        item = self.tableWidget.item(current_row, 0)
        if not item:
            return
        entry = item.data(Qt.ItemDataRole.UserRole) or {}
        urlText = f'source URL: {entry.get("url", "")}' if entry.get("type") == "source" else f'binary URL: {entry.get("binary_urls", "")}'
        
        # Get version info
        name = entry.get("name", "unknown")
        pdk_dir = self.pdksDir / re.sub(r"[^A-Za-z0-9_.-]", "_", name)
        config_path = pdk_dir / "config.json"
        version_info = f"Remote Version: {entry.get('version', 'N/A')}"
        
        if config_path.exists():
            try:
                with open(config_path) as f:
                    config = json.load(f)
                    local_version = config.get("version", "Unknown")
                    version_info += f"\nInstalled Version: {local_version}"
                    if self._compare_versions(local_version, entry.get('version', '0.0.0')) < 0:
                        version_info += " (Update available)"
            except Exception:
                pass
        
        text = f"{entry.get('description', '')}\n\nProcess: {entry.get('process', 'N/A')}\n{version_info}\nLicense: {entry.get('license', 'Unknown')}\n{urlText}"
        self.desc.setPlainText(text)

    def _on_item_activated(self, item: QTableWidgetItem):
        entry = item.data(Qt.ItemDataRole.UserRole) or {}
        name = entry.get("name", "pdk")
        ret = QMessageBox.question(
            self,
            "Install PDK",
            f"Install PDK '{name}'?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        )
        if ret == QMessageBox.StandardButton.Yes:
            self._install_entry(entry)

    def _on_download(self):
        current_row = self.tableWidget.currentRow()
        if current_row < 0:
            QMessageBox.information(self, "No selection", "Please select a PDK first.")
            return
        item = self.tableWidget.item(current_row, 0)
        if item:
            entry = item.data(Qt.ItemDataRole.UserRole) or {}
            name = entry.get("name", "pdk")
            ret = QMessageBox.question(
                self,
                "Install PDK",
                f"Install '{name}'?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
            if ret == QMessageBox.StandardButton.Yes:
                self._install_entry(entry)

    def _install_entry(self, entry: dict):
        name = re.sub(r"[^A-Za-z0-9_.-]", "_", entry.get("name", "pdk"))
        url = (
            self._get_binary_url(entry)
            if entry.get("type") == "binary"
            else entry.get("url")
        )

        if not url:
            QMessageBox.warning(self, "Error", "No URL for your platform.")
            return

        target_subdir = self.pdksDir / name
        if target_subdir.exists():
            ok = QMessageBox.question(
                self,
                "Overwrite",
                f"{name} exists. Overwrite?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
            if ok != QMessageBox.StandardButton.Yes:
                return
            shutil.rmtree(target_subdir, ignore_errors=True)

        try:
            with urllib.request.urlopen(url) as resp:
                tmp_fd, tmp_path = tempfile.mkstemp()
                os.close(tmp_fd)
                with open(tmp_path, "wb") as out:
                    out.write(resp.read())

            self.pdksDir.mkdir(parents=True, exist_ok=True)
            if url.lower().endswith(".zip"):
                with zipfile.ZipFile(tmp_path, "r") as zf:
                    zf.extractall(path=self.pdksDir)
            else:
                target_subdir.mkdir(parents=True, exist_ok=True)
                (target_subdir / Path(url).name).write_bytes(
                    Path(tmp_path).read_bytes()
                )

            os.remove(tmp_path)
            self.fetch_registry()
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e))

    def _get_binary_url(self, entry: dict) -> str:
        binary_urls = entry.get("binary_urls", {})

        system = platform.system().lower()
        arch = platform.machine().lower()
        py_ver = f"py{sys.version_info.major}{sys.version_info.minor}"

        # Try most specific first
        for key in [f"{system}-{arch}-{py_ver}", f"{system}-{arch}", system]:
            if key in binary_urls:
                return binary_urls[key]
        return ""


    def _on_uninstall(self):
        current_row = self.tableWidget.currentRow()
        if current_row < 0:
            QMessageBox.information(self, "No selection", "Please select a PDK first.")
            return
        item = self.tableWidget.item(current_row, 0)
        if not item:
            return
        entry = item.data(Qt.ItemDataRole.UserRole) or {}
        name = entry.get("name", "pdk")
        pdk_dir = self.pdksDir / re.sub(r"[^A-Za-z0-9_.-]", "_", name)

        if not pdk_dir.exists():
            QMessageBox.information(
                self, "Not Installed", f"'{name}' is not installed."
            )
            return

        ret = QMessageBox.question(
            self,
            "Uninstall",
            f"Uninstall '{name}'?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        )
        if ret == QMessageBox.StandardButton.Yes:
            self._uninstall_pdk(re.sub(r"[^A-Za-z0-9_.-]", "_", name), pdk_dir)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

# This module includes all the base definitions for schematic drawings.
from dataclasses import replace

from PySide6.QtCore import (Qt)
from PySide6.QtGui import (QColor, QPen, QBrush)

import revedaEditor.backend.dataDefinitions as ddef

# schematic layers

wireLayer = ddef.edLayer(
    name="wire",
    pcolor=QColor("cyan"),
    pwidth=2,
    pstyle=Qt.SolidLine,
    z=0,
    bcolor=QColor("cyan"),
    bstyle=Qt.SolidPattern,
    visible=True,
    selectable=True,
)
wireErrorLayer = replace(wireLayer, name="wireError", pcolor=QColor("red"),
                         bcolor=QColor("red"), z=1)
selectedWireLayer = replace(wireLayer, name="selectedWire", pcolor=QColor("blue"), bcolor=
QColor("blue"), z=2)
wireHilightLayer = ddef.edLayer(
    name="wireHilightLayer",
    pcolor=QColor("darkMagenta"),
    pwidth=5,
    z=6,
    bcolor=QColor("darkMagenta"),
    bstyle=Qt.SolidPattern,
    visible=True,
    selectable=False,
)
guideLineLayer = replace(
    wireLayer, name="guideLine", pcolor=QColor("gray"), pstyle=Qt.DashLine, z=7
)
textLayer = ddef.edLayer(
    name="text", pcolor=QColor("white"), pwidth=1, z=4, visible=True, selectable=True
)

schematicPinLayer = ddef.edLayer(
    name="schematicPin",
    pcolor=QColor("red"),
    pwidth=1,
    z=3,
    bcolor=QColor("red"),
    bstyle=Qt.SolidPattern,
    visible=True,
    selectable=True,
)
selectedSchematicPinLayer = replace(
    schematicPinLayer, name="selectedSchematicPin", pcolor=QColor("yellow"), z=4
)
selectedTextLayer = replace(
    textLayer, name="selectedText", pcolor=QColor("yellow"), z=5
)
schematicPinNameLayer = replace(schematicPinLayer, name="schematicPin", pcolor=QColor(
    'yellow'), bcolor=QColor('yellow'))

ignoreSymbolLayer = ddef.edLayer(
    name="ignoreLayer",
    pcolor=QColor("red"),
    pwidth=5,
    z=6,
    visible=True,
    selectable=False,
)
otherLayer = ddef.edLayer(
    name="otherLayer",
    pcolor=QColor("gray"),
    bcolor=QColor("gray"),
    pwidth=1,
    z=0,
    visible=True,
    selectable=False,
)

draftLayer = ddef.edLayer(
    name="draftLayer",
    pcolor=QColor("gray"),
    pwidth=2,
    z=10,
    bcolor=QColor("gray"),
    bstyle=Qt.DiagCrossPattern,
    visible=True,
    selectable=True,
)
# schematic pens
schematicPinPen = QPen(
    schematicPinLayer.pcolor, schematicPinLayer.pwidth, schematicPinLayer.pstyle
)
selectedSchematicPinPen = QPen(
    selectedSchematicPinLayer.pcolor,
    selectedSchematicPinLayer.pwidth,
    selectedSchematicPinLayer.pstyle,
)
textPen = QPen(textLayer.pcolor, textLayer.pwidth, textLayer.pstyle)
selectedTextPen = QPen(
    selectedTextLayer.pcolor, selectedTextLayer.pwidth, selectedTextLayer.pstyle
)
guideLinePen = QPen(guideLineLayer.pcolor, guideLineLayer.pwidth, guideLineLayer.pstyle)
guideLinePen.setCosmetic(True)
wirePen = QPen(wireLayer.pcolor, wireLayer.pwidth, wireLayer.pstyle)

selectedWirePen = QPen(
    selectedWireLayer.pcolor, selectedWireLayer.pwidth, selectedWireLayer.pstyle
)
stretchWirePen = QPen(QColor("red"), wireLayer.pwidth, wireLayer.pstyle)
errorWirePen = QPen(wireErrorLayer.pcolor, wireErrorLayer.pwidth, wireErrorLayer.pstyle)
ignoreSymbolPen = QPen(
    ignoreSymbolLayer.pcolor, ignoreSymbolLayer.pwidth, ignoreSymbolLayer.pstyle
)
hilightPen = QPen(
    wireHilightLayer.pcolor, wireHilightLayer.pwidth, wireHilightLayer.pstyle
)
otherPen = QPen(otherLayer.pcolor, otherLayer.pwidth, otherLayer.pstyle)
draftPen = QPen(draftLayer.pcolor, draftLayer.pwidth, draftLayer.pstyle)
draftPen.setCosmetic(True)

# schematic brushes
schematicPinBrush = QBrush(schematicPinLayer.bcolor, schematicPinLayer.bstyle)
wireBrush = QBrush(wireLayer.bcolor, wireLayer.bstyle)
selectedWireBrush = QBrush(selectedWireLayer.bcolor, selectedWireLayer.bstyle)
errorWireBrush = QBrush(wireErrorLayer.bcolor, wireErrorLayer.bstyle)
selectedSchematicPinBrush = QBrush(
    selectedSchematicPinLayer.bcolor, selectedSchematicPinLayer.bstyle
)
otherBrush = QBrush(otherLayer.bcolor, otherLayer.bstyle)
draftBrush = QBrush(draftLayer.bcolor, draftLayer.bstyle)

# crossing dot diameter
crossingDotDiameter = 2



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service
#    whose value derives, entirely or substantially, from the functionality of the Software.
#    Any license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


import json
import logging
import pathlib
import shutil
from typing import List

from PySide6.QtCore import Qt, QPoint
from PySide6.QtGui import (QAction, QStandardItemModel, QStandardItem, )
from PySide6.QtWidgets import (QAbstractItemView, QDialog, QMenu, QMessageBox, QTreeView,
                               QWidget, QApplication, QListView, QHBoxLayout, QVBoxLayout,
                               QLabel, )

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.libBackEnd as libb
import revedaEditor.backend.libraryMethods as libm
import revedaEditor.gui.fileDialogues as fd
import revedaEditor.gui.layoutDialogues as ldlg
import revedaEditor.gui.textEditor as ted
from revedaEditor.gui.configEditor import configEditor


class BaseDesignLibrariesView(QWidget):
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.parent = parent
        self._app = QApplication.instance()
        self.libBrowsW = self.parentWidget().parentWidget()
        self.appMainW = self.libBrowsW.appMainW
        self.libraryDict = self.appMainW.libraryDict  # type: dict
        self.cellViews = self.libBrowsW.cellViews  # type: list
        self.openViews = self.appMainW.openViews  # type: dict
        self.logger = logging.getLogger("reveda")

        # Common selection tracking
        self.selectedLib = None
        self.selectedCell = None
        self.selectedView = None

        # Initialize the model (to be implemented by child classes)
        self.libraryModel = designLibrariesModel(self.libraryDict)

    def removeLibrary(self, selectedLib: libb.libraryItem):
        try:
            button = QMessageBox.question(self, "Library Deletion",
                                          "Are you sure to delete this library? This action cannot be undone.", )
            if button == QMessageBox.Yes:
                self.libraryModel.removeLibraryFromModel(selectedLib)
                self.libraryDict.pop(selectedLib.libraryName, None)
                self.reworkDesignLibrariesView(self.libraryDict)
                self.libBrowsW.writeLibDefFile(self.libraryDict, self.libBrowsW.libFilePath)
        except Exception as e:
            self.logger.error(f"Error removing library: {e}")

    def renameLib(self, selectedLib: libb.libraryItem):
        try:
            oldLibraryName = selectedLib.libraryName
            dlg = fd.renameLibDialog(self, oldLibraryName)
            if dlg.exec() == QDialog.DialogCode.Accepted:
                newLibraryName = dlg.newLibraryName.text().strip()
                libraryItem = libm.getLibItem(self.libraryModel, oldLibraryName)
                oldLibraryPath = libraryItem.data(Qt.ItemDataRole.UserRole + 2)
                self.libraryModel.removeRow(libraryItem.row())
                newLibraryPath = oldLibraryPath.parent.joinpath(newLibraryName)
                oldLibraryPath.rename(newLibraryPath)
                self.libraryDict.pop(oldLibraryName)
                self.libraryDict[newLibraryName] = pathlib.Path(newLibraryPath)
                self.reworkDesignLibrariesView(self.libraryDict)
                self.libBrowsW.writeLibDefFile(self.libraryDict, self.libBrowsW.libFilePath)
        except Exception as e:
            self.logger.error(f"Error renaming library: {e}")

    def openView(self, selectedViewItem: libb.viewItem):
        try:
            cellItem = selectedViewItem.parent()
            libItem = cellItem.parent()
            viewItemT = ddef.viewItemTuple(libItem, cellItem, selectedViewItem)
            self.openCellView(viewItemT)
        except Exception as e:
            self.logger.error(f"Error opening view: {e}")

    def createNewCellView(self, itemTuple: ddef.viewItemTuple):
        """
            this method is used to open the editor after a viewitem is created.
        """
        viewNameT = itemTuple.convertToViewNameTuple()
        if itemTuple.viewItem.viewType in ["schematic", "symbol", "layout"]:
            from revedaEditor.gui.editorFactory import EditorFactory
            editor = EditorFactory.createEditor(itemTuple.viewItem.viewType, itemTuple.viewItem,
                                                self.libraryDict,
                                                self)
            self.appMainW.openViews[viewNameT] = editor
            if hasattr(editor, "loadSchematic"):
                editor.loadSchematic()
            elif hasattr(editor, "loadSymbol"):
                editor.loadSymbol()
            elif hasattr(editor, "loadLayout"):
                editor.loadLayout()
            editor.show()
        elif itemTuple.viewItem.viewType == "veriloga":
            self._handle_veriloga_view(itemTuple)
        elif itemTuple.viewItem.viewType == "spice":
            self._handle_spice_view(itemTuple)
        elif itemTuple.viewItem.viewType == "config":
            schViewsList = [itemTuple.cellItem.child(row).viewName for row in
                            range(itemTuple.cellItem.rowCount()) if
                            itemTuple.cellItem.child(row).viewType == "schematic"]
            dlg = fd.createConfigViewDialogue(self.appMainW)
            dlg.libraryNameEdit.setText(itemTuple.libraryItem.libraryName)
            dlg.cellNameEdit.setText(itemTuple.cellItem.cellName)
            dlg.viewNameCB.addItems(schViewsList)
            dlg.switchViews.setText(", ".join(self.appMainW.switchViewList))
            dlg.stopViews.setText(", ".join(self.appMainW.stopViewList))
            if dlg.exec() == QDialog.DialogCode.Accepted:
                from revedaEditor.gui.configEditor import createNewConfigView
                configWindow = createNewConfigView(itemTuple.cellItem, itemTuple.viewItem, dlg,
                                                   self.libraryDict,
                                                   self)
                self.appMainW.openViews[viewNameT] = configWindow
                configWindow.show()
        elif itemTuple.viewItem.viewType == "pcell":
            dlg = ldlg.pcellLinkDialogue(self.appMainW, itemTuple.viewItem)
            if dlg.exec() == QDialog.DialogCode.Accepted:
                items = [{"cellView": "pcell"}, {"reference": dlg.pcellCB.currentText()}]
                with itemTuple.viewItem.data(Qt.ItemDataRole.UserRole + 2).open(
                        mode="w+") as pcellFile:
                    json.dump(items, pcellFile, indent=4)
            else:
                try:
                    itemTuple.viewItem.data(Qt.ItemDataRole.UserRole + 2).unlink()
                    itemTuple.viewItem.parent().removeRow(itemTuple.viewItem.row())
                except OSError as e:
                    self.logger.warning(f"Error:{e.strerror}")
        else:
            if hasattr(self._app, 'pluginsObj'):
                self._app.pluginsObj.createCellView(itemTuple)

    def reworkDesignLibrariesView(self, libraryDict: dict):
        """
        To be implemented by child classes.
        """
        pass

    def _handle_text_view(self, viewItemT: ddef.viewItemTuple, editor_class, finished_callback):
        """Handle text-based view opening (spice/veriloga)."""
        viewNameT = viewItemT.convertToViewNameTuple()
        filePath = None
        with open(viewItemT.viewItem.viewPath) as f:
            try:
                filePath = pathlib.Path(viewItemT.cellItem.cellPath.joinpath(json.load(f)[1].get("filePath")))
            except IndexError:
                if viewItemT.viewItem.viewType == 'spice':
                    filePath = pathlib.Path(viewItemT.cellItem.cellPath.joinpath(
                        viewNameT.cellName).with_suffix(".sp"))
                    filePath.touch(exist_ok=True)
                elif viewItemT.viewItem.viewType == 'veriloga':
                    filePath = pathlib.Path(viewItemT.cellItem.cellPath.joinpath(
                        viewNameT.cellName).with_suffix(".va"))
                    filePath.touch(exist_ok=True)
        
        if filePath:
            filePath = filePath.resolve()
            editor = editor_class(filePath)
            editor.cellViewTuple = viewNameT
            editor.closedSignal.connect(finished_callback)
            self.appMainW.openViews[viewNameT] = editor
            editor.show()
        else:
            self.logger.warning(f"File {filePath} does not exist")


    def _handle_spice_view(self, viewItemT: ddef.viewItemTuple):
        """Handle spice view opening."""
        self._handle_text_view(viewItemT, ted.spiceEditor, self.spiceEditFinished)

    def _handle_veriloga_view(self, viewItemT: ddef.viewItemTuple):
        """Handle veriloga view opening."""
        self._handle_text_view(viewItemT, ted.verilogaEditor, self.verilogaEditFinished)

    def openCellView(self, viewItemT: ddef.viewItemTuple):
        from revedaEditor.gui.editorFactory import EditorFactory
        viewNameT = viewItemT.convertToViewNameTuple()

        if viewNameT in self.appMainW.openViews.keys():
            self.appMainW.openViews[viewNameT].show()
            return viewNameT

        view_type = viewItemT.viewItem.viewType

        # Handle standard editor types using factory
        if view_type in EditorFactory.getSupportedViewTypes():
            editor = EditorFactory.createEditor(view_type, viewItemT.viewItem,
                                                self.libraryDict,
                                                self)

            # Load content based on editor type - use view_type instead of hasattr
            if view_type == 'layout':
                editor.loadLayout()  # type: ignore[attr-defined]
            elif view_type == 'schematic':
                editor.loadSchematic()  # type: ignore[attr-defined]
            elif view_type == 'symbol':
                editor.loadSymbol()  # type: ignore[attr-defined]

            editor.show()

            # Fit items in view if available
            if hasattr(editor, 'centralW') and hasattr(editor.centralW, 'scene'):
                editor.centralW.scene.fitItemsInView()
            self.appMainW.openViews[viewNameT] = editor

        # Handle special cases that require custom logic
        elif view_type == "config":
            editor = configEditor(viewItemT.viewItem, self.libraryDict,
                                  self)
            editor.loadConfig()
        elif view_type == "spice":
            self._handle_spice_view(viewItemT)
        elif view_type == "veriloga":
            self._handle_veriloga_view(viewItemT)
        else:
            if hasattr(self._app, 'pluginsObj'):
                result = self._app.pluginsObj.openCellView(viewItemT)
                if not result:
                    self.logger.warning(f"No handler for view type: {view_type}")

        return viewNameT

    def verilogaEditFinished(self, editor: ted.verilogaEditor):
        import revedaEditor.fileio.importVeriloga as imva
        imva.importVerilogaModule(editor.cellViewTuple, str(editor.filePathObj))
        # self.appMainW.openViews.pop(editor.cellViewTuple)

    def spiceEditFinished(self, editor: ted.spiceEditor):
        import revedaEditor.fileio.importSpice as imsp
        imsp.importSpiceSubckt(editor.cellViewTuple, str(editor.filePathObj))
        # self.appMainW.openViews.pop(editor.cellViewTuple)


class designLibrariesColumnView(BaseDesignLibrariesView):
    def __init__(self, parent):
        super().__init__(parent=parent)  # QTreeView

        # Create three list views with labels
        self.libsListView = QListView()

        self.libsListView.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.libsListView.customContextMenuRequested.connect(self.libsListContextMenuEvent)
        self.cellsListView = QListView()
        self.cellsListView.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.cellsListView.customContextMenuRequested.connect(
            self.cellsListContextMenuEvent)
        self.viewsListView = QListView()
        self.viewsListView.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
        self.viewsListView.customContextMenuRequested.connect(
            self.viewsListContextMenuEvent)

        libsLabel = QLabel("**Libraries**")
        libsLabel.setTextFormat(Qt.MarkdownText)
        cellsLabel = QLabel("**Cells**")
        cellsLabel.setTextFormat(Qt.MarkdownText)
        viewsLabel = QLabel("**Cell Views**")
        viewsLabel.setTextFormat(Qt.MarkdownText)
        # Create a horizontal layout and add the list views to it
        layout = QHBoxLayout()

        # Create vertical layouts for each list with its label
        libsLayout = QVBoxLayout()
        libsLayout.addWidget(libsLabel)
        libsLayout.addWidget(self.libsListView)

        cellsLayout = QVBoxLayout()
        cellsLayout.addWidget(cellsLabel)
        cellsLayout.addWidget(self.cellsListView)

        viewsLayout = QVBoxLayout()
        viewsLayout.addWidget(viewsLabel)
        viewsLayout.addWidget(self.viewsListView)

        # Add layouts to main layout
        layout.addLayout(libsLayout)
        layout.addLayout(cellsLayout)
        layout.addLayout(viewsLayout)
        self.setLayout(layout)

        self.libsListView.setModel(self.libraryModel)
        # Connect selection signals
        self.libsListView.selectionModel().selectionChanged.connect(
            self.onLibsListSelection)

    def onLibsListSelection(self, selected, deselected):
        # Clear second and third lists
        self.cellsListView.setModel(None)
        self.viewsListView.setModel(None)

        # Get the selected index
        indexes = selected.indexes()
        if not indexes:
            return

        # Create new model for second list
        cellsModel = QStandardItemModel()
        cellsModel.setHorizontalHeaderLabels(["Cells"])
        cellsModel.setSortRole(Qt.ItemDataRole.UserRole + 3)

        # Get the selected item and its children
        self.selectedLib = self.libraryModel.itemFromIndex(indexes[0])

        children = [self.selectedLib.child(i) for i in range(self.selectedLib.rowCount())]
        if self.selectedLib and self.selectedLib.hasChildren():
            for cellItem in children:
                clonedCellItem = cellItem.clone()
                cellsModel.appendRow(clonedCellItem)

        self.cellsListView.setModel(cellsModel)
        # Connect second list selection after setting its model
        self.cellsListView.selectionModel().selectionChanged.connect(
            self.onCellsListSelection)

    def recursive_clone(self, item):
        """Recursively clone an item and all its children."""
        clonedItem = item.clone()
        # Store reference to original item
        clonedItem.setData(item,
                           Qt.ItemDataRole.UserRole + 10)  # Use a custom role to store the original item

        if item.hasChildren():
            for i in range(item.rowCount()):
                child = item.child(i)
                clonedChild = self.recursive_clone(child)
                clonedItem.appendRow(clonedChild)
        return clonedItem

    def onCellsListSelection(self, selected, deselected):
        # Clear third list
        self.viewsListView.setModel(None)

        # Get the selected index
        indexes = selected.indexes()
        if not indexes:
            return
        cellItem = (
            self.cellsListView.model().itemFromIndex(indexes[0]).data(
                Qt.ItemDataRole.UserRole + 10))
        # Create new model for third list
        viewsModel = self.createViewsListModel(cellItem=cellItem)
        self.viewsListView.setModel(viewsModel)
        # Connect third list selection after setting its model
        self.viewsListView.selectionModel().selectionChanged.connect(
            self.onViewsListSelection)

    def createViewsListModel(self, cellItem: libb.cellItem) -> QStandardItemModel:
        """
        Create a new model for the views list based on the selected cell item.
        """
        viewsModel = QStandardItemModel()
        viewsModel.setHorizontalHeaderLabels(["Cell Views"])
        viewsModel.setSortRole(Qt.ItemDataRole.UserRole + 3)

        if cellItem and cellItem.hasChildren():
            for i in range(cellItem.rowCount()):
                child = cellItem.child(i)
                cloned_child = child.clone()
                cloned_child.setData(child, Qt.ItemDataRole.UserRole + 10)
                viewsModel.appendRow(cloned_child)
        return viewsModel

    def onViewsListSelection(self, selected, deselected):
        indexes = selected.indexes()
        if indexes:
            self.selectedView = self.viewsListView.model().itemFromIndex(indexes[0])

    def reworkDesignLibrariesView(self, libraryDict: dict):
        """
        Recreate library model from libraryDict.
        """
        # Disconnect existing selection signals
        try:
            self.libsListView.selectionModel().selectionChanged.disconnect()
        except:
            pass

        # Create new model and set it
        self.libraryModel = designLibrariesModel(libraryDict)
        self.libsListView.setModel(self.libraryModel)

        # Reconnect selection signals
        self.libsListView.selectionModel().selectionChanged.connect(
            self.onLibsListSelection)

        # Clear other views
        self.cellsListView.setModel(None)
        self.viewsListView.setModel(None)

        # Update library model reference in parent
        self.libBrowsW.libraryModel = self.libraryModel

    def libsListContextMenuEvent(self, pos: QPoint):
        senderView = self.sender()
        menu = QMenu()
        index = senderView.indexAt(pos)
        if index.isValid():
            selectedLibItem = self.libsListView.model().itemFromIndex(index)
            menu.addAction(QAction("Rename Library", self,
                                   triggered=lambda: self.renameLib(selectedLibItem), ))
            menu.addAction(QAction("Remove Library", self,
                                   triggered=lambda: self.removeLibrary(selectedLibItem), ))
            menu.addAction(QAction("Create Cell", self,
                                   triggered=lambda: self.createCell(selectedLibItem), ))
            menu.addAction(QAction("File Information...", self,
                                   triggered=lambda: self.showItemFileInfo(
                                       selectedLibItem), ))
            menu.exec(senderView.viewport().mapToGlobal(
                pos))  # Use global position for context menu

    def cellsListContextMenuEvent(self, pos):
        senderView = self.sender()
        menu = QMenu()
        index = senderView.indexAt(pos)
        if index.isValid():
            selectedCloneCellItem = self.cellsListView.model().itemFromIndex(index)
            selectedCellItem = selectedCloneCellItem.data(Qt.ItemDataRole.UserRole + 10)
            menu.addAction(QAction("Create CellView...", self,
                                   triggered=lambda: self.createCellView(
                                       selectedCloneCellItem), ))
            menu.addAction(QAction("Copy Cell...", self,
                                   triggered=lambda: self.copyCell(selectedCellItem), ))
            menu.addAction(QAction("Rename Cell...", self,
                                   triggered=lambda: self.renameCell(
                                       selectedCloneCellItem), ))
            menu.addAction(QAction("Delete Cell...", self,
                                   triggered=lambda: self.deleteCell(
                                       selectedCloneCellItem), ))
            menu.addAction(QAction("File Information...", self,
                                   triggered=lambda: self.showClonedItemFileInfo(
                                       selectedCloneCellItem), ))
            menu.exec(senderView.viewport().mapToGlobal(
                pos))  # Use global position for context menu

    def viewsListContextMenuEvent(self, pos):
        senderView = self.sender()
        menu = QMenu()
        index = senderView.indexAt(pos)
        if index.isValid():
            selectedCloneViewItem = self.viewsListView.model().itemFromIndex(index)
            selectedViewItem = selectedCloneViewItem.data(Qt.ItemDataRole.UserRole + 10)
            menu.addAction(QAction("Open View", self,
                                   triggered=lambda: self.openView(selectedViewItem)))
            menu.addAction(QAction("Copy View...", self, triggered=lambda: self.copyView(
                selectedCloneViewItem), ))
            menu.addAction(QAction("Rename View...", self,
                                   triggered=lambda: self.renameView(
                                       selectedCloneViewItem), ))
            menu.addAction(QAction("Delete View...", self,
                                   triggered=lambda: self.deleteView(
                                       selectedCloneViewItem), ))
            menu.addAction(QAction("File Information...", self,
                                   triggered=lambda: self.showClonedItemFileInfo(
                                       selectedCloneViewItem), ))
            menu.exec(senderView.viewport().mapToGlobal(
                pos))  # Use global position for context menu

    def createCell(self, selectedLib: libb.libraryItem):
        try:
            dlg = fd.createCellDialog(self, self.libraryModel)
            dlg.libNamesCB.setCurrentText(selectedLib.libraryName)
            if dlg.exec() == QDialog.DialogCode.Accepted:
                cellName = dlg.cellCB.currentText()
                if cellName.strip() != "":
                    selectedLib = libm.getLibItem(self.libraryModel,
                                                  dlg.libNamesCB.currentText())
                    newCellItem = libb.createCell(self, selectedLib, cellName)
                    if newCellItem:
                        # add now a clone of newCellItem to temporary cellListModel
                        cloneItem = newCellItem.clone()
                        self.cellsListView.model().appendRow(cloneItem)
                else:
                    self.logger.error("Please enter a cell name.")
        except OSError as e:
            self.logger.warning(f"Error creating cell: {e}")

    def copyCell(self, selectedCellItem: libb.cellItem):
        try:
            parentLib: libb.libraryItem = selectedCellItem.parent()
            dlg = fd.copyCellDialog(self)
            dlg.libraryCB.setModel(self.libraryModel)
            dlg.libraryCB.setCurrentText(parentLib.libraryName)
            pass
            if dlg.exec() == QDialog.DialogCode.Accepted:
                success, newCellItem = libb.copyCell(self, self.libraryModel,
                                                     selectedCellItem, dlg.copyName.text(),
                                                     dlg.selectedLibPath, )
                if success:
                    cloneItem = newCellItem.clone()
                    if selectedCellItem.hasChildren():
                        for i in range(selectedCellItem.rowCount()):
                            child = selectedCellItem.child(i)
                            clonedChild = child.clone()
                            cloneItem.appendRow(clonedChild)
                    self.cellsListView.model().appendRow(cloneItem)
                else:
                    self.logger.error("Failed to copy cell.")
        except OSError as e:
            self.logger.warning(f"Error copying cell: {e}")

    def renameCell(self, cloneCellItem: libb.cellItem):
        try:
            oldName = cloneCellItem.cellName
            cellItem = cloneCellItem.data(Qt.ItemDataRole.UserRole + 10)
            dlg = fd.renameCellDialog(self, cellItem)
            libName = cellItem.parent().libraryName
            if dlg.exec() == QDialog.DialogCode.Accepted:
                newName = dlg.nameEdit.text().strip()
                libb.renameCell(self, cloneCellItem, newName)
                # update the original cell item
                libb.renameCell(self, cellItem, dlg.nameEdit.text().strip(), )
                updateJSONFieldInCell(self.libraryModel, libName, 'cell', oldName, newName)
                self.logger.info(f"Renamed {oldName} to {newName}")
        except OSError as e:
            self.logger.warning(f"Error renaming cell: {e}")

    def deleteCell(self, selectedCloneCellItem: libb.cellItem):
        try:
            originalCell = selectedCloneCellItem.data(Qt.ItemDataRole.UserRole + 10)
            shutil.rmtree(selectedCloneCellItem.data(Qt.ItemDataRole.UserRole + 2))
            self.libraryModel.removeRow(originalCell.row())
            if self.cellsListView.model():
                self.cellsListView.model().removeRow(selectedCloneCellItem.row())
            self.logger.info(f"Cell {originalCell.cellName} deleted.")
            self.reworkDesignLibrariesView(self.libraryModel.libraryDict)

        except OSError as e:
            self.logger.warning(f"Error deleting cell: {e}")

    def copyView(self, selectedCloneViewItem: libb.viewItem):
        selectedViewItem = selectedCloneViewItem.data(Qt.ItemDataRole.UserRole + 10)
        dlg = fd.copyViewDialog(self, self.libraryModel)
        dlg.libNamesCB.setCurrentText(selectedViewItem.parent().parent().libraryName)
        dlg.cellCB.setCurrentText(selectedViewItem.parent().cellName)
        if dlg.exec() != QDialog.DialogCode.Accepted:
            return

        if selectedViewItem.data(Qt.ItemDataRole.UserRole + 1) != "view":
            self.logger.error("Selected item is not a view.")
            return

        viewPath = selectedViewItem.data(Qt.ItemDataRole.UserRole + 2)
        libName = dlg.libNamesCB.currentText()
        cellName = dlg.cellCB.currentText()
        newViewName = dlg.viewName.text().strip()

        selectedLibItem = libm.getLibItem(self.libraryModel, libName)
        if not selectedLibItem:
            self.logger.error("Selected library not found.")
            return

        # Find or create cell
        cellItem = libm.getCellItem(selectedLibItem, cellName)  # noqa: F811
        if not cellItem:
            cellItem = libb.createCell(self.libBrowsW,  selectedLibItem,
                                       cellName)

        # Check if view already exists
        if any(child.viewName == newViewName for child in
               (cellItem.child(row) for row in range(cellItem.rowCount()))):
            self.logger.warning("View already exists. Delete cellview and try again.")
            return

        newViewPath = cellItem.data(Qt.ItemDataRole.UserRole + 2).joinpath(
            f"{newViewName}.json")
        try:
            newViewPath.parent.mkdir(parents=True, exist_ok=True)
            newViewItem = libb.viewItem(newViewPath)
            shutil.copy(viewPath, newViewPath)
            cellItem.appendRow(newViewItem)
            viewsModel = self.createViewsListModel(cellItem)
            self.viewsListView.setModel(viewsModel)
            self.logger.info(f"View {newViewName} copied successfully.")
        except Exception as e:
            self.logger.error(f"Failed to copy view: {e}")

    def createCellView(self, selectedCloneCellItem: libb.cellItem):
        cellItem = selectedCloneCellItem.data(Qt.ItemDataRole.UserRole + 10)
        dlg = fd.newCellViewDialog(self, self.libraryModel)
        dlg.libNamesCB.setCurrentText(cellItem.parent().libraryName)
        dlg.cellCB.setCurrentText(cellItem.cellName)
        dlg.viewType.addItems(self.libBrowsW.cellViews)
        if dlg.exec() == QDialog.DialogCode.Accepted:
            self.handleNewCellView(cellItem, dlg)

    def handleNewCellView(self, cellItem, dlg):
        viewName = dlg.viewName.text().strip()
        libItem = libm.getLibItem(self.libraryModel, dlg.libNamesCB.currentText())
        viewItem = libm.findViewItem(self.libraryModel, libItem.libraryName,
                                     cellItem.cellName, viewName)

        if viewItem:
            itemTuple = ddef.viewItemTuple(libItem, cellItem, viewItem)
            messagebox = QMessageBox(self)
            messagebox.setText("Cell view already exists.")
            messagebox.setIcon(QMessageBox.Warning)
            messagebox.setWindowTitle(f"{viewItem.viewName} already exists")
            messagebox.setStandardButtons(QMessageBox.Save | QMessageBox.Discard)
            messagebox.setDefaultButton(QMessageBox.Discard)
            result = messagebox.exec()
            if result == QMessageBox.Save:
                self.viewsListView.model().removeRow(viewItem.row())
                viewItem = libb.createCellView(self.appMainW, viewName, cellItem)
                self.createNewCellView(itemTuple)
                # Add the new view item to the views list
                cloneViewItem = viewItem.clone()
                self.viewsListView.model().appendRow(cloneViewItem)
        else:
            viewItem = libb.createCellView(self.appMainW, viewName, cellItem)
            itemTuple = ddef.viewItemTuple(libItem, cellItem, viewItem)
            # cellItem.appendRow(viewItem)
            self.createNewCellView(itemTuple)
            # Add the new view item to the views list
            cloneViewItem = viewItem.clone()
            self.viewsListView.model().appendRow(cloneViewItem)

    def renameView(self, selectedCloneViewItem: libb.viewItem):
        selectedViewItem = selectedCloneViewItem.data(Qt.ItemDataRole.UserRole + 10)
        oldViewName = selectedViewItem.viewName
        cellItem = selectedViewItem.parent()  # noqa: F811
        dlg = fd.renameViewDialog(self.libBrowsW, oldViewName)
        if dlg.exec() == QDialog.DialogCode.Accepted:
            newName = dlg.newViewNameEdit.text()
            try:
                viewPathObj = selectedViewItem.data(Qt.ItemDataRole.UserRole + 2)
                newPathObj = viewPathObj.parent.joinpath(f"{newName}.json")
                if newPathObj.exists():
                    raise FileExistsError
                viewPathObj.rename(newPathObj)
                # Update the view item with the new path and name

                newViewItem = libb.viewItem(newPathObj)
                cellItem.appendRow(newViewItem)
                cellItem.removeRow(selectedViewItem.row())
                viewsModel = self.createViewsListModel(cellItem)
                self.viewsListView.setModel(viewsModel)
                self.logger.info(f"View {oldViewName} renamed to {newName}.")
            except FileExistsError:
                self.logger.error("Cellview exists.")

    def deleteView(self, selectedCloneViewItem: libb.viewItem):
        try:
            selectedCloneViewItem.data(Qt.ItemDataRole.UserRole + 2).unlink()
            selectedViewItem = selectedCloneViewItem.data(Qt.ItemDataRole.UserRole + 10)
            # Remove the original item from the model
            itemRow = selectedViewItem.row()
            cellItem = selectedViewItem.parent()
            cellItem.removeRow(itemRow)
            # Remove the cloned item from the view
            viewsModel = self.createViewsListModel(cellItem)
            self.viewsListView.setModel(viewsModel)
            self.logger.info(f"View {selectedViewItem.viewName} deleted.")
            # self.reworkDesignLibrariesView(self.libraryModel.libraryDict)
        except FileNotFoundError:
            self.logger.warning("View file not found.")
        except PermissionError:
            self.logger.warning("Permission denied while deleting view.")

        except Exception as e:
            self.logger.warning(f"Error:{e}")

    def showClonedItemFileInfo(self, selectedCloneItem: libb.viewItem):
        selectedItem = selectedCloneItem.data(Qt.ItemDataRole.UserRole + 10)
        viewPath = selectedItem.data(Qt.ItemDataRole.UserRole + 2)
        if viewPath.exists():
            dlg = fd.fileInfoDialogue(viewPath, self.libBrowsW)
        dlg.exec()

    def showItemFileInfo(self, selectedItem: libb.viewItem):
        viewPath = selectedItem.data(Qt.ItemDataRole.UserRole + 2)
        if viewPath.exists():
            dlg = fd.fileInfoDialogue(viewPath, self.libBrowsW)
        dlg.exec()


class designLibrariesTreeView(BaseDesignLibrariesView):
    """
    Deprecated for the moment. 4/3/2026.
    """
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.libraryModel.setSortRole(Qt.ItemDataRole.UserRole + 3)
        self.treeView = QTreeView()
        self.treeView.setModel(self.libraryModel)
        self.treeView.setSelectionMode(QAbstractItemView.SingleSelection)
        self.treeView.setUniformRowHeights(True)
        # self.treeView.expandAll()
        layout = QVBoxLayout()
        layout.addWidget(self.treeView)
        self.setLayout(layout)

    def createCell(self, selectedLib: libb.libraryItem):
        try:
            dlg = fd.createCellDialog(self, self.libraryModel)
            dlg.libNamesCB.setCurrentText(selectedLib.libraryName)
            if dlg.exec() == QDialog.DialogCode.Accepted:
                cellName = dlg.cellCB.currentText()
                if cellName.strip() != "":
                    libb.createCell(self, selectedLib, cellName)
                else:
                    self.logger.error("Please enter a cell name.")
        except OSError as e:
            self.logger.warning(f"Error in creating cell:{e}")

    def copyCell(self, selectedCellItem: libb.cellItem):
        try:
            parentLibrary = selectedCellItem.parent()
            dlg = fd.copyCellDialog(self)
            dlg.libraryCB.setCurrentText(parentLibrary.libraryName)
            if dlg.exec() == QDialog.DialogCode.Accepted:
                success, newCellItem = libb.copyCell(self, self.libraryModel,
                                                     selectedCellItem, dlg.copyName.text(),
                                                     dlg.selectedLibPath, )
                if success:
                    self.logger.info("Cell copied successfully.")

        except OSError as e:
            self.logger.warning(f"Error in copying cell:{e}")

    def renameCell(self, selectedCell: libb.cellItem):
        try:
            dlg = fd.renameCellDialog(self, selectedCell)
            if dlg.exec() == QDialog.DialogCode.Accepted:
                oldName = selectedCell.cellName
                libName = selectedCell.parent().libraryName
                success = libb.renameCell(self, dlg.cellItem, dlg.nameEdit.text())
                if success:
                    newName = selectedCell.cellName
                    self.logger.info(f"Cell {oldName} renamed to {dlg.nameEdit.text()}.")
                    updateJSONFieldInCell(self.libraryModel, libName, 'cell', oldName,
                                          newName)
        except OSError as e:
            self.logger.warning(f"Error in renaming cell:{e}")

    def deleteCell(self, selectedCell: libb.cellItem):
        try:
            shutil.rmtree(selectedCell.data(Qt.ItemDataRole.UserRole + 2))
            self.libraryModel.removeRow(selectedCell.row())
            self.logger.info(f"Cell {selectedCell.cellName} deleted.")
        except OSError as e:
            self.logger.warning(f"Error in deleting cell:{e}")

    def createCellView(self, selectedCell: libb.cellItem):
        try:
            dlg = fd.newCellViewDialog(self, self.libraryModel)
            dlg.libNamesCB.setCurrentText(selectedCell.parent().libraryName)
            dlg.cellCB.setCurrentText(selectedCell.cellName)
            dlg.viewType.addItems(self.libBrowsW.cellViews)
            if dlg.exec() == QDialog.DialogCode.Accepted:
                self.handleNewCellView(selectedCell, dlg)
        except OSError as e:
            self.logger.warning(f"Error in creating cell view:{e}")

    def handleNewCellView(self, cellItem:libb.cellItem, dlg):
        libItem:libb.libraryItem = cellItem.parent()
        viewName = dlg.viewName.text().strip()
        viewItem = libm.findViewItem(self.libraryModel, libItem.libraryName,
                                     cellItem.cellName, viewName)
        itemTuple = ddef.viewItemTuple(libItem, cellItem, viewItem)
        if viewItem:
            messagebox = QMessageBox(self)
            messagebox.setText("Cell view already exists.")
            messagebox.setIcon(QMessageBox.Warning)
            messagebox.setWindowTitle(f"{viewItem.viewName} already exists")
            messagebox.setStandardButtons(QMessageBox.Save | QMessageBox.Discard)
            messagebox.setDefaultButton(QMessageBox.Discard)
            result = messagebox.exec()
            if result == QMessageBox.Save:
                cellItem.removeRow(viewItem.row())
                viewItem = libb.createCellView(self.libBrowsW, viewName, cellItem)
                cellItem.appendRow(viewItem)
                self.createNewCellView(itemTuple)

        else:
            viewItem = libb.createCellView(self.appMainW, viewName.strip(), cellItem)
            cellItem.appendRow(viewItem)
            self.createNewCellView(itemTuple)

    def copyView(self, selectedView: libb.viewItem):
        try:
            dlg = fd.copyViewDialog(self, self.libraryModel)
            dlg.libNamesCB.setCurrentText(selectedView.parent().parent().libraryName)
            dlg.cellCB.setCurrentText(selectedView.parent().cellName)
            if dlg.exec() == QDialog.DialogCode.Accepted:
                if selectedView.data(Qt.ItemDataRole.UserRole + 1) == "view":
                    viewPath = self.selectedItem.data(Qt.ItemDataRole.UserRole + 2)
                    selectedLibItem = libm.getLibItem(self.libraryModel,
                                                      dlg.libNamesCB.currentText())
                    cellName = dlg.cellCB.currentText()
                    libCellNames = [selectedLibItem.child(row).cellName for row in
                                    range(selectedLibItem.rowCount())]
                    if (
                            cellName in libCellNames):  # check if there is the cell in the library
                        cellItem = libm.getCellItem(selectedLibItem,
                                                    dlg.cellCB.currentText())
                    else:
                        cellItem = libb.createCell(self.libBrowsW, self.libraryModel,
                                                   selectedLibItem,
                                                   dlg.cellCB.currentText(), )
                    cellViewNames = [cellItem.child(row).viewName for row in
                                     range(cellItem.rowCount())]
                    newViewName = dlg.viewName.text()
                    if newViewName in cellViewNames:
                        self.logger.warning(
                            "View already exists. Delete cellview and try again.")
                    else:
                        newViewPath = cellItem.data(Qt.ItemDataRole.UserRole + 2).joinpath(
                            f"{newViewName}.json")
                        shutil.copy(viewPath, newViewPath)
                        cellItem.appendRow(libb.viewItem(newViewPath))
        except OSError as e:
            self.logger.warning(f"Error in copying view:{e}")

    def renameView(self, selectedView: libb.viewItem):
        try:
            oldViewName = selectedView.viewName
            dlg = fd.renameViewDialog(self.libBrowsW, oldViewName)
            if dlg.exec() == QDialog.DialogCode.Accepted:
                newName = dlg.newViewNameEdit.text()
                try:
                    viewPathObj = selectedView.data(Qt.ItemDataRole.UserRole + 2)
                    newPathObj = selectedView.data(Qt.ItemDataRole.UserRole + 2).rename(
                        viewPathObj.parent.joinpath(f"{newName}.json"))
                    selectedView.parent().appendRow(libb.viewItem(newPathObj))
                    selectedView.parent().removeRow(selectedView.row())
                except FileExistsError:
                    self.logger.error("Cellview exists.")
        except OSError as e:
            self.logger.warning(f"Error in renaming view:{e}")

    def deleteView(self, selectedView: libb.viewItem):
        try:
            selectedView.data(Qt.ItemDataRole.UserRole + 2).unlink()
            itemRow = selectedView.row()
            parent = selectedView.parent()
            parent.removeRow(itemRow)
        except OSError as e:
            self.logger.warning(
                f"Error in removing item: {selectedView.viewName}:{e.strerror}")

    def reworkDesignLibrariesView(self, libraryDict: dict):
        """
        Recreate library model from libraryDict.
        """
        self.libraryModel = designLibrariesModel(libraryDict)
        self.setModel(self.libraryModel)
        self.libBrowsW.libraryModel = self.libraryModel

    def showFileInfo(self, selectedItem: libb.viewItem):
        viewPath = selectedItem.data(Qt.ItemDataRole.UserRole + 2)
        if viewPath.exists():
            dlg = fd.fileInfoDialogue(viewPath, self.libBrowsW)
        dlg.exec()

    # context menu
    # def contextMenuEvent(self, event):
    #     menu = QMenu(self)
    #     pos = event.pos()
    #     # Use self directly instead of self.sender()
    #     index = self.treeView.indexAt(pos)
    #     if index.isValid():
    #         selectedItem = self.libraryModel.itemFromIndex(index)
    def contextMenuEvent(self, event):
        menu = QMenu(self)
        pos = self.treeView.mapFromGlobal(event.globalPos())
        index = self.treeView.indexAt(pos)
        if index.isValid():
            selectedItem = self.libraryModel.itemFromIndex(index)
            if selectedItem.data(Qt.ItemDataRole.UserRole + 1) == "library":
                menu.addAction(QAction("Rename Library", self.treeView,
                                       triggered=lambda: self.renameLib(selectedItem), ))
                menu.addAction(QAction("Remove Library", self.treeView,
                                       triggered=lambda: self.removeLibrary(
                                           selectedItem), ))
                menu.addAction(QAction("Create Cell", self.treeView,
                                       triggered=lambda: self.createCell(selectedItem), ))
            elif selectedItem.data(Qt.ItemDataRole.UserRole + 1) == "cell":
                menu.addAction(QAction("Create CellView...", self,
                                       triggered=lambda: self.createCellView(
                                           selectedItem), ))
                menu.addAction(QAction("Copy Cell...", self.treeView, triggered=lambda: (
                    self.copyCell(selectedItem.parent())), ))
                menu.addAction(QAction("Rename Cell...", self.treeView,
                                       triggered=lambda: self.renameCell(selectedItem), ))
                menu.addAction(QAction("Delete Cell...", self.treeView,
                                       triggered=lambda: self.deleteCell(selectedItem), ))
            elif selectedItem.data(Qt.ItemDataRole.UserRole + 1) == "view":
                menu.addAction(QAction("Open View", self.treeView,
                                       triggered=lambda: self.openView(selectedItem), ))
                menu.addAction(QAction("Copy View...", self,
                                       triggered=lambda: self.copyView(selectedItem), ))
                menu.addAction(QAction("Rename View...", self.treeView,
                                       triggered=lambda: self.renameView(selectedItem), ))
                menu.addAction(QAction("Delete View...", self.treeView,
                                       triggered=lambda: self.deleteView(selectedItem), ))
            menu.addAction(QAction("File Information...", self.treeView,
                                   triggered=lambda: self.showFileInfo(selectedItem), ))
            # Use global position for context menu
            menu.exec(event.globalPos())


class designLibrariesModel(QStandardItemModel):
    def __init__(self, libraryDict):
        self.libraryDict = libraryDict
        super().__init__()

        self.setHorizontalHeaderLabels(["Libraries"])
        self.initModel()
        self.logger = logging.getLogger("reveda")

    def initModel(self):
        for designPath in self.libraryDict.values():
            self.populateLibrary(designPath)

    def populateLibrary(self, designPath: pathlib.Path) -> None:  # designPath: Path
        """
        Populate library view.
        """
        if designPath.joinpath("reveda.lib").exists():
            libraryItem = self.addLibraryToModel(designPath)
            cellList = [cell.name for cell in designPath.iterdir() if cell.is_dir()]
            for cell in cellList:
                cellItem = self.addCellToModel(designPath.joinpath(cell), libraryItem)
                viewList = [view.name for view in designPath.joinpath(cell).iterdir() if
                            view.suffix == ".json"]
                for view in viewList:
                    self.addViewToModel(designPath.joinpath(cell, view), cellItem)
                cellItem.sortChildren(0)
            libraryItem.sortChildren(0)

    def addLibraryToModel(self, designPath: pathlib.Path) -> libb.libraryItem:
        libraryEntry = libb.libraryItem(designPath)
        for row in range(self.invisibleRootItem().rowCount()):
            existingItem = self.invisibleRootItem().child(row)
            if existingItem.data(Qt.ItemDataRole.UserRole + 2) == designPath:
                self.logger.warning(f"Library {designPath} already exists in the model.")
                break
        else:
            self.invisibleRootItem().appendRow(libraryEntry)
        return libraryEntry

    def removeLibraryFromModel(self, libraryItem: libb.libraryItem) -> None:
        shutil.rmtree(libraryItem.data(Qt.ItemDataRole.UserRole + 2), ignore_errors=True)
        self.invisibleRootItem().removeRow(libraryItem.row())

    def addCellToModel(self, cellPath: pathlib.Path,
                       parentItem: libb.libraryItem) -> libb.cellItem:
        cellEntry = libb.cellItem(cellPath)
        parentItem.appendRow(cellEntry)
        return cellEntry

    def addViewToModel(self, viewPath: pathlib.Path,
                       parentItem: libb.cellItem) -> libb.viewItem:
        viewEntry = libb.viewItem(viewPath)
        parentItem.appendRow(viewEntry)
        return viewEntry

    def listLibraries(self) -> List[str]:
        librariesList = []
        for row in range(self.rowCount()):
            itemText = self.item(row, 0).text()
            if itemText:
                librariesList.append(itemText)
        return librariesList

    def listLibraryCells(self, libraryName: str) -> List[str]:
        cellsList = []
        libraryItem = libm.getLibItem(self, libraryName)
        if libraryItem:
            for row in range(libraryItem.rowCount()):
                itemText = libraryItem.child(row, 0).text()
                if itemText:
                    cellsList.append(itemText)
        return cellsList

    def listCellViews(self, libraryName: str, cellName: str, viewTypes: List[str]) -> List[
        str]:
        viewsList = []
        libraryItem = libm.getLibItem(self, libraryName)
        cellItem = libm.getCellItem(libraryItem, cellName)
        if cellItem:
            for row in range(cellItem.rowCount()):
                if cellItem.child(row, 0).viewType in viewTypes:
                    viewsList.append(cellItem.child(row, 0).text())
        return viewsList


class symbolViewsModel(designLibrariesModel):
    """
    Initializes the object with the given `libraryDict` and `symbolViews`.

    Parameters:
        libraryDict (dict): A dictionary containing the library information.
        symbolViews (list): A list of symbol views.

    Returns:
        None
    """

    def __init__(self, libraryDict: dict, symbolViews: list):
        self.symbolViews = symbolViews
        super().__init__(libraryDict)

    def populateLibrary(self, designPath):  # designPath: Path
        """
        Populate library view.
        """
        if designPath.joinpath("reveda.lib").exists():
            libraryItem = self.addLibraryToModel(designPath)
            cellList = [cell.name for cell in designPath.iterdir() if cell.is_dir()]
            for cell in cellList:
                cellItem = self.addCellToModel(designPath.joinpath(cell), libraryItem)
                viewList = [view.name for view in designPath.joinpath(cell).iterdir() if
                            view.suffix == ".json" and any(
                                x in view.name for x in self.symbolViews)]
                for view in viewList:
                    self.addViewToModel(designPath.joinpath(cell, view), cellItem)


class layoutViewsModel(designLibrariesModel):
    def __init__(self, libraryDict: dict, layoutViews: list):
        self.layoutViews = layoutViews
        super().__init__(libraryDict)

    def populateLibrary(self, designPath):  # designPath: Path
        """
        Populate library view.
        """
        if designPath.joinpath("reveda.lib").exists():
            libraryItem = self.addLibraryToModel(designPath)
            cellList = [cell.name for cell in designPath.iterdir() if cell.is_dir()]
            for cell in cellList:
                cellItem = self.addCellToModel(designPath.joinpath(cell), libraryItem)
                viewList = [view.name for view in designPath.joinpath(cell).iterdir() if
                            view.suffix == ".json" and any(
                                x in view.name for x in self.layoutViews)]
                for view in viewList:
                    self.addViewToModel(designPath.joinpath(cell, view), cellItem)


class libraryCheckListView(QListView):
    def __init__(self, parent, model: designLibrariesModel):
        super().__init__(parent)
        self.designLibrariesModel = model
        self.setWindowTitle("Library Check List")
        self.setGeometry(100, 100, 400, 600)
        self.model = QStandardItemModel(self)
        self.setModel(self.model)

        libraries = self.designLibrariesModel.listLibraries()
        for library in libraries:
            item = QStandardItem(library)
            item.setCheckable(True)
            item.setCheckState(Qt.Unchecked)
            self.model.appendRow(item)

    def getCheckedLibraries(self):
        checkedLibraries = []
        for row in range(self.model.rowCount()):
            item = self.model.item(row)
            if item.checkState() == Qt.Checked:
                checkedLibraries.append(item.text())
        return checkedLibraries


def updateJSONFieldInLibrary(model: "designLibrariesModel", libraryName: str, key: str,
                             oldValue: str, newValue: str) -> None:
    """
    Update a specific JSON field in all view files within a single library.

    Args:
        model: The designLibrariesModel instance
        libraryName: Name of the library to process
        key: The JSON key to search for
        oldValue: The old value to match
        newValue: The new value to set for the key
    """
    import json
    libItem = libm.getLibItem(model, libraryName)
    if libItem.hasChildren():
        for row in range(libItem.rowCount()):
            cellItem = libItem.child(row)
            if cellItem.hasChildren():
                for row in range(cellItem.rowCount()):
                    viewItem = cellItem.child(row)
                    try:
                        with open(viewItem.viewPath, "r") as f:
                            data = json.load(f)
                            updated = False
                            for item in data:
                                if item.get(key) == oldValue:
                                    updated = True
                                    item[key] = newValue

                        if updated:
                            with open(viewItem.viewPath, "w") as f:
                                json.dump(data, f, indent=4)
                    except Exception as e:
                        print(f"Error updating {viewItem.viewPath}: {str(e)}")


def updateJSONFieldInCell(model: "designLibrariesModel", libraryName: str, cellName: str,
                          key: str, oldValue: str, newValue: str) -> None:
    """
    Update a specific JSON field in all view files within a single cell.

    Args:
        model: The designLibrariesModel instance
        libraryName: Name of the library to process
        cellName: Name of the cell to process
        key: The JSON key to search for
        oldValue: The old value to match
        newValue: The new value to set for the key
    """
    import json
    libItem = libm.getLibItem(model, libraryName)
    cellItem = libm.getCellItem(libItem, cellName)

    if cellItem.hasChildren():
        for row in range(cellItem.rowCount()):
            viewItem = cellItem.child(row)
            try:
                with open(viewItem.viewPath, "r") as f:
                    data = json.load(f)
                    updated = False
                    for item in data:
                        if item.get(key) == oldValue:
                            updated = True
                            item[key] = newValue

                if updated:
                    with open(viewItem.viewPath, "w") as f:
                        json.dump(data, f, indent=4)
            except Exception as e:
                print(f"Error updating {viewItem.viewPath}: {str(e)}")



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

# properties dialogues for various editor functions

from PySide6.QtWidgets import (
    QDialog,
    QVBoxLayout,
    QFormLayout,
    QDialogButtonBox,
    QLabel,
    QComboBox,
)


class findProjectEditors(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Find Related Editors")
        self.setMinimumWidth(300)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.fLayout.setContentsMargins(10, 20, 10, 20)
        self.relatedEditorsCB = QComboBox()
        self.fLayout.addRow(QLabel("Find Related Editors:"), self.relatedEditorsCB)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()



#    "Commons Clause" License Condition v1.0
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

import json
import os
import platform
import re
import shutil
import sys
import tempfile
import urllib.request
import zipfile
from pathlib import Path

from PySide6.QtCore import Qt
from PySide6.QtWidgets import (
    QHBoxLayout,
    QLabel,
    QTableWidget,
    QTableWidgetItem,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QProgressBar,
    QTextEdit,
    QVBoxLayout,
    QWidget,
    QHeaderView,
)


class PluginRegistryWindow(QMainWindow):
    DEFAULT_REGISTRY = "https://raw.githubusercontent.com/eskiyerli/revolutionEDA_plugins/main/plugins.json"

    def __init__(
        self,
        parent=None,
        registry_url: str | None = None,
        plugins_dir: Path | None = None,
    ):
        super().__init__(parent)
        self.setWindowTitle("Revolution EDA Plugin Registry")
        self.resize(800, 400)

        self.registry_url = registry_url or self.DEFAULT_REGISTRY
        self.pluginsDir = (
            Path(os.environ.get("REVEDA_PLUGIN_PATH"))
            if os.environ.get("REVEDA_PLUGIN_PATH")
            else (plugins_dir or Path.cwd() / "plugins")
        )
        self.pluginsDir = self.pluginsDir.resolve()

        self._initUI()
        self._registry = []
        self.fetch_registry()

    def _initUI(self):
        central = QWidget()
        self.setCentralWidget(central)
        main_l = QHBoxLayout(central)

        left_w = QWidget()
        left_l = QVBoxLayout(left_w)
        left_l.setContentsMargins(0, 0, 0, 0)
        self.tableWidget = QTableWidget()
        self.tableWidget.setColumnCount(5)
        self.tableWidget.setHorizontalHeaderLabels(
            ["Installed", "Plugin", "Type", "Version", "License"]
        )
        self.tableWidget.horizontalHeader().setSectionResizeMode(
            QHeaderView.ResizeMode.Stretch
        )
        self.tableWidget.setSelectionBehavior(QTableWidget.SelectionBehavior.SelectRows)
        self.tableWidget.setSelectionMode(QTableWidget.SelectionMode.SingleSelection)
        left_l.addWidget(QLabel("Available plugins:"))
        left_l.addWidget(self.tableWidget)
        self.refresh_btn = QPushButton("Refresh")
        left_l.addWidget(self.refresh_btn)
        main_l.addWidget(left_w, 2)

        right_w = QWidget()
        right_l = QVBoxLayout(right_w)
        right_l.setContentsMargins(0, 0, 0, 0)
        right_l.addWidget(QLabel("Description:"))
        self.desc = QTextEdit()
        self.desc.setReadOnly(True)
        right_l.addWidget(self.desc, 1)
        self.download_btn = QPushButton("Download / Install")
        self.uninstall_btn = QPushButton("Uninstall")
        self.progress = QProgressBar()
        self.progress.setRange(0, 100)
        right_l.addWidget(self.progress)
        right_l.addWidget(self.download_btn)
        right_l.addWidget(self.uninstall_btn)
        main_l.addWidget(right_w, 2)

        self.tableWidget.itemSelectionChanged.connect(self._onSelect)
        self.tableWidget.itemDoubleClicked.connect(self._on_item_activated)
        self.download_btn.clicked.connect(self._on_download)
        self.uninstall_btn.clicked.connect(self._on_uninstall)
        self.refresh_btn.clicked.connect(self.fetch_registry)

    def fetch_registry(self):
        self.tableWidget.setRowCount(0)
        self.desc.clear()
        self.progress.setValue(0)
        try:
            with urllib.request.urlopen(self.registry_url) as resp:
                raw = resp.read()
                index = json.loads(raw.decode("utf-8"))
                if isinstance(index, dict) and "plugins" in index:
                    index = index["plugins"]
                self._registry = index if isinstance(index, list) else []
        except Exception as exc:
            QMessageBox.warning(
                self, "Registry Error", f"Failed to fetch registry:\n{exc}"
            )
            self._registry = []

        self.tableWidget.setRowCount(len(self._registry))
        for row, entry in enumerate(self._registry):
            name = entry.get("name", "unknown")
            plugin_dir = self.pluginsDir / re.sub(r"[^A-Za-z0-9_.-]", "_", name)

            checkbox_item = QTableWidgetItem()
            checkbox_item.setFlags(Qt.ItemFlag.ItemIsEnabled)
            checkbox_item.setCheckState(
                Qt.CheckState.Checked
                if plugin_dir.exists()
                else Qt.CheckState.Unchecked
            )
            checkbox_item.setData(Qt.ItemDataRole.UserRole, entry)
            self.tableWidget.setItem(row, 0, checkbox_item)

            self.tableWidget.setItem(row, 1, QTableWidgetItem(name))
            self.tableWidget.setItem(
                row, 2, QTableWidgetItem(entry.get("type", "source").title())
            )
            self.tableWidget.setItem(
                row, 3, QTableWidgetItem(entry.get("version", "0.0.0"))
            )
            self.tableWidget.setItem(
                row, 4, QTableWidgetItem(entry.get("license", "Unknown"))
            )

    def _onSelect(self):
        self.desc.clear()
        current_row = self.tableWidget.currentRow()
        if current_row < 0:
            return
        item = self.tableWidget.item(current_row, 0)
        if not item:
            return
        entry = item.data(Qt.ItemDataRole.UserRole) or {}
        url_text = ""
        if entry.get('type') == 'binary':
            url_text = f"Binary URLs: {entry.get('binary_urls', {})}"
        else:
            url_text = f"URL: {entry.get('url', '')}"
        text = f"{entry.get('description', '')}\n\nType: {entry.get('type', 'source').title()}\nVersion: {entry.get('version', 'N/A')}\nLicense: {entry.get('license', 'Unknown')}\n{url_text}"
        self.desc.setPlainText(text)

    def _on_item_activated(self, item: QTableWidgetItem):
        entry = item.data(Qt.ItemDataRole.UserRole) or {}
        name = entry.get("name", "plugin")
        ret = QMessageBox.question(
            self,
            "Install Plugin",
            f"Install plugin '{name}'?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        )
        if ret == QMessageBox.StandardButton.Yes:
            self._install_entry(entry)

    def _on_download(self):
        current_row = self.tableWidget.currentRow()
        if current_row < 0:
            QMessageBox.information(
                self, "No selection", "Please select a plugin first."
            )
            return
        item = self.tableWidget.item(current_row, 0)
        if item:
            entry = item.data(Qt.ItemDataRole.UserRole) or {}
            name = entry.get("name", "plugin")
            ret = QMessageBox.question(
                self,
                "Install Plugin",
                f"Install '{name}'?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
            if ret == QMessageBox.StandardButton.Yes:
                self._install_entry(entry)

    def _install_entry(self, entry: dict):
        name = re.sub(r"[^A-Za-z0-9_.-]", "_", entry.get("name", "plugin"))
        url = (
            self._get_binary_url(entry)
            if entry.get("type") == "binary"
            else entry.get("url")
        )

        if not url:
            QMessageBox.warning(self, "Error", "No URL for your platform.")
            return

        target_subdir = self.pluginsDir / name
        if target_subdir.exists():
            ok = QMessageBox.question(
                self,
                "Overwrite",
                f"{name} exists. Overwrite?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
            )
            if ok != QMessageBox.StandardButton.Yes:
                return
            shutil.rmtree(target_subdir, ignore_errors=True)

        try:
            with urllib.request.urlopen(url) as resp:
                tmp_fd, tmp_path = tempfile.mkstemp()
                os.close(tmp_fd)
                with open(tmp_path, "wb") as out:
                    out.write(resp.read())

            self.pluginsDir.mkdir(parents=True, exist_ok=True)
            if url.lower().endswith(".zip"):
                with zipfile.ZipFile(tmp_path, "r") as zf:
                    zf.extractall(path=self.pluginsDir)
            else:
                target_subdir.mkdir(parents=True, exist_ok=True)
                (target_subdir / Path(url).name).write_bytes(
                    Path(tmp_path).read_bytes()
                )

            os.remove(tmp_path)
            self.fetch_registry()
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e))

    def _get_binary_url(self, entry: dict) -> str:
        binary_urls = entry.get("binary_urls", {})
        if not binary_urls:
            return entry.get("url")

        system = platform.system().lower()
        arch = platform.machine().lower()
        py_ver = f"py{sys.version_info.major}{sys.version_info.minor}"

        # Try most specific first
        for key in [f"{system}-{arch}-{py_ver}", f"{system}-{arch}", system]:
            if key in binary_urls:
                return binary_urls[key]

        return entry.get("url")

    def _on_uninstall(self):
        current_row = self.tableWidget.currentRow()
        if current_row < 0:
            QMessageBox.information(
                self, "No selection", "Please select a plugin first."
            )
            return
        item = self.tableWidget.item(current_row, 0)
        if not item:
            return
        entry = item.data(Qt.ItemDataRole.UserRole) or {}
        name = entry.get("name", "plugin")
        plugin_dir = self.pluginsDir / re.sub(r"[^A-Za-z0-9_.-]", "_", name)

        if not plugin_dir.exists():
            QMessageBox.information(
                self, "Not Installed", f"'{name}' is not installed."
            )
            return

        ret = QMessageBox.question(
            self,
            "Uninstall",
            f"Uninstall '{name}'?",
            QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
        )
        if ret == QMessageBox.StandardButton.Yes:
            try:
                shutil.rmtree(plugin_dir)
                self.fetch_registry()
            except Exception as e:
                QMessageBox.critical(self, "Error", str(e))



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)
#

from pathlib import Path
from typing import Dict, TYPE_CHECKING

from PySide6.QtCore import (
    Qt,
)

from PySide6.QtGui import (QWindow, QStandardItem, QFontDatabase, QDoubleValidator)
from PySide6.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QRadioButton,
    QButtonGroup,
    QGroupBox,
    QWidget,
    QCheckBox,
    QTableWidget,
    QTableWidgetItem,
)

import revedaEditor.backend.drcModelView as drcmv
import revedaEditor.common.layoutShapes as lshp
import revedaEditor.gui.editFunctions as edf
from revedaEditor.backend.pdkLoader import importPDKModule

# from dotenv import load_dotenv

process = importPDKModule("process")
if TYPE_CHECKING:
    pass


class layoutInstanceDialogue(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Layout Instance/Pcell Options")
        self.setMinimumWidth(400)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        vLayout = QVBoxLayout()

        instanceParamsGroup = QGroupBox("Instance Parameters")
        self.instanceParamsLayout = QFormLayout()
        instanceParamsGroup.setLayout(self.instanceParamsLayout)
        self.instanceLibName = edf.longLineEdit()
        self.instanceParamsLayout.addRow("Library:", self.instanceLibName)
        self.instanceCellName = edf.longLineEdit()
        self.instanceParamsLayout.addRow("Cell:", self.instanceCellName)
        self.instanceViewName = edf.longLineEdit()
        # self.pinstanceViewName.setReadOnly(True)
        self.instanceParamsLayout.addRow("View:", self.instanceViewName)
        vLayout.addWidget(instanceParamsGroup)

        self.pcellParamsGroup = QGroupBox("Parametric Cell Parameters")
        self.pcellParamsLayout = QFormLayout()
        self.pcellParamsGroup.setLayout(self.pcellParamsLayout)
        vLayout.addWidget(self.pcellParamsGroup)
        self.pcellParamsGroup.hide()

        self.locationGroup = QGroupBox("Location")
        self.locationLayout = QFormLayout()
        self.locationGroup.setLayout(self.locationLayout)
        self.xEdit = edf.shortLineEdit()
        self.yEdit = edf.shortLineEdit()
        self.locationLayout.addRow("Location X:", self.xEdit)
        self.locationLayout.addRow("Location Y:", self.yEdit)
        vLayout.addWidget(self.locationGroup)
        self.locationGroup.hide()
        vLayout.addWidget(self.buttonBox)
        self.setLayout(vLayout)
        self.show()


class layoutInstancePropertiesDialogue(layoutInstanceDialogue):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("PCell Instance Properties")
        self.instanceNameEdit = edf.longLineEdit()
        self.instanceParamsLayout.addRow("Instance Name:", self.instanceNameEdit)
        self.locationGroup.show()


class pcellLinkDialogue(QDialog):
    def __init__(self, parent, viewItem: QStandardItem):
        super().__init__(parent)
        # self.logger = parentW.logger
        self.viewItem = viewItem
        # self.pcells = self.getClasses()
        self.pcells = list(importPDKModule("pcells").pcells.keys())
        self.setWindowTitle("PCell Settings")
        self.setMinimumSize(400, 200)
        self.mainLayout = QVBoxLayout()
        groupBox = QGroupBox()
        groupLayout = QVBoxLayout()
        formLayout = QFormLayout()
        groupBox.setLayout(groupLayout)
        self.pcellCB = QComboBox()
        self.pcellCB.addItems(self.pcells)
        formLayout.addRow(edf.boldLabel("PCell:"), self.pcellCB)
        groupLayout.addLayout(formLayout)
        self.mainLayout.addWidget(groupBox)
        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()

    # @staticmethod
    # def getClasses():
    #     module = importPDKModule('pcells')
    #     return [name for name, obj in inspect.getmembers(module, inspect.isclass)
    #             if issubclass(obj, module.baseCell)]


class createPathDialogue(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Create Path")
        # self.setMinimumSize(300, 300)
        mainLayout = QVBoxLayout()
        self.pathOrientBox = QGroupBox("Path Orientation")
        horizontalLayout = QHBoxLayout(self.pathOrientBox)
        self.pathOrientBox.setLayout(horizontalLayout)
        pathOrientGroup = QButtonGroup()
        self.manhattanButton = QRadioButton("Manhattan")
        pathOrientGroup.addButton(self.manhattanButton)
        horizontalLayout.addWidget(self.manhattanButton)
        self.diagonalButton = QRadioButton("Diagonal")
        pathOrientGroup.addButton(self.diagonalButton)
        horizontalLayout.addWidget(self.diagonalButton)
        self.anyButton = QRadioButton("Any")
        pathOrientGroup.addButton(self.anyButton)
        horizontalLayout.addWidget(self.anyButton)
        self.horizontalButton = QRadioButton("Horizontal")
        pathOrientGroup.addButton(self.horizontalButton)
        horizontalLayout.addWidget(self.horizontalButton)
        self.verticalButton = QRadioButton("Vertical")
        pathOrientGroup.addButton(self.verticalButton)
        horizontalLayout.addWidget(self.verticalButton)
        self.manhattanButton.setChecked(True)
        pathOrientGroup.setExclusive(True)
        mainLayout.addWidget(self.pathOrientBox)
        groupBox = QGroupBox()
        self.formLayout = QFormLayout()
        groupBox.setLayout(self.formLayout)
        self.pathLayerCB = QComboBox()
        self.formLayout.addRow(edf.boldLabel("Path Layer:"), self.pathLayerCB)
        self.pathWidth = edf.shortLineEdit()
        self.pathWidthValidator = QDoubleValidator(self)
        self.pathWidth.setValidator(self.pathWidthValidator)
        self.formLayout.addRow(edf.boldLabel("Path Width:"), self.pathWidth)
        self.pathNameEdit = edf.shortLineEdit()
        self.formLayout.addRow(edf.boldLabel("Path Name:"), self.pathNameEdit)
        self.startExtendEdit = edf.shortLineEdit()
        self.formLayout.addRow(edf.boldLabel("Start Extend:"), self.startExtendEdit)
        self.endExtendEdit = edf.shortLineEdit()
        self.formLayout.addRow(edf.boldLabel("End Extend:"), self.endExtendEdit)
        mainLayout.addWidget(groupBox)

        QBtn = QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)
        self.show()


class layoutPathPropertiesDialog(createPathDialogue):
    def __init__(self, parent) -> None:
        super().__init__(parent)
        self.setWindowTitle("Path Properties")
        # self.mainLayout.removeWidget(self.pathOrientBox)
        self.p1PointEditX = edf.shortLineEdit()
        self.p1PointEditY = edf.shortLineEdit()
        self.p2PointEditX = edf.shortLineEdit()
        self.p2PointEditY = edf.shortLineEdit()
        self.angleEdit = edf.shortLineEdit()
        self.formLayout.addRow(edf.boldLabel("P1 Point X:"), self.p1PointEditX)
        self.formLayout.addRow(edf.boldLabel("P1 Point Y:"), self.p1PointEditY)
        self.formLayout.addRow(edf.boldLabel("P2 Point X:"), self.p2PointEditX)
        self.formLayout.addRow(edf.boldLabel("P2 Point Y:"), self.p2PointEditY)
        self.formLayout.addRow(edf.boldLabel("Path Angle:"), self.angleEdit)


class createLayoutPinDialog(QDialog):
    def __init__(self, parent) -> None:
        super().__init__(parent)
        self.setWindowTitle("Create Layout Pin")
        self.setMinimumWidth(300)
        fontFamilies = QFontDatabase.families(QFontDatabase.WritingSystem.Latin)
        fixedFamilies = [
            family for family in fontFamilies if QFontDatabase.isFixedPitch(family)
        ]
        QBtn = (
                QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )

        self.mainLayout = QVBoxLayout()
        self.pinPropGroupBox = QGroupBox("Pin Properties")
        fLayout = QFormLayout()
        self.pinPropGroupBox.setLayout(fLayout)
        self.pinName = QLineEdit()
        self.pinName.setPlaceholderText("Pin Name")
        self.pinName.setToolTip("Enter pin name")
        fLayout.addRow(edf.boldLabel("Pin Name"), self.pinName)
        self.pinDir = QComboBox()
        self.pinDir.addItems(lshp.layoutPin.pinDirs)
        self.pinDir.setToolTip("Select pin direction")
        fLayout.addRow(edf.boldLabel("Pin Direction"), self.pinDir)
        self.pinType = QComboBox()
        self.pinType.addItems(lshp.layoutPin.pinTypes)
        self.pinType.setToolTip("Select pin type")
        fLayout.addRow(edf.boldLabel("Pin Type"), self.pinType)
        self.mainLayout.addWidget(self.pinPropGroupBox)
        self.layerSelectGroupBox = QGroupBox("Select layers")
        self.layerFormLayout = QFormLayout()
        self.layerSelectGroupBox.setLayout(self.layerFormLayout)
        self.pinLayerCB = QComboBox()
        self.layerFormLayout.addRow(edf.boldLabel("Pin Layer:"), self.pinLayerCB)
        self.labelLayerCB = QComboBox()
        self.labelLayerText = edf.boldLabel("Label Layer:")
        self.layerFormLayout.addRow(self.labelLayerText, self.labelLayerCB)
        self.mainLayout.addWidget(self.layerSelectGroupBox)
        labelPropBox = QGroupBox("Label Properties")
        labelPropLayout = QFormLayout()
        labelPropBox.setLayout(labelPropLayout)
        self.familyCB = QComboBox()
        self.familyCB.addItems(fixedFamilies)
        self.familyCB.currentTextChanged.connect(self.familyFontStyles)
        labelPropLayout.addRow(edf.boldLabel("Font Name"), self.familyCB)
        self.fontStyleCB = QComboBox()
        self.fontStyles = QFontDatabase.styles(fixedFamilies[0])
        self.fontStyleCB.addItems(self.fontStyles)
        self.fontStyleCB.currentTextChanged.connect(self.styleFontSizes)
        labelPropLayout.addRow(edf.boldLabel("Font Style"), self.fontStyleCB)
        self.labelHeightCB = QComboBox()
        self.fontSizes = [
            str(size)
            for size in QFontDatabase.pointSizes(fixedFamilies[0], self.fontStyles[0])
        ]
        self.labelHeightCB.addItems(self.fontSizes)
        labelPropLayout.addRow(edf.boldLabel("Label Height"), self.labelHeightCB)
        self.labelAlignCB = QComboBox()
        self.labelAlignCB.addItems(lshp.layoutLabel.LABEL_ALIGNMENTS)
        labelPropLayout.addRow(QLabel("Label Alignment"), self.labelAlignCB)
        self.labelOrientCB = QComboBox()
        self.labelOrientCB.addItems(lshp.layoutLabel.LABEL_ORIENTS)
        labelPropLayout.addRow(QLabel("Label Orientation"), self.labelOrientCB)
        self.mainLayout.addWidget(labelPropBox)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()

    def familyFontStyles(self, s):
        self.fontStyleCB.clear()
        self.fontStyles = QFontDatabase.styles(self.familyCB.currentText())
        self.fontStyleCB.addItems(self.fontStyles)

    def styleFontSizes(self, s):
        self.labelHeightCB.clear()
        selectedFamily = self.familyCB.currentText()
        selectedStyle = self.fontStyleCB.currentText()
        self.fontSizes = [
            str(size)
            for size in QFontDatabase.pointSizes(selectedFamily, selectedStyle)
        ]
        self.labelHeightCB.addItems(self.fontSizes)


class layoutPinProperties(QDialog):
    def __init__(self, parent: QWidget):
        super().__init__(parent)

        self.setWindowTitle("Layout Pin Properties")
        self.setMinimumWidth(300)

        QBtn = (
                QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )

        self.mainLayout = QVBoxLayout()
        pinPropGroupBox = QGroupBox("Pin Properties")
        fLayout = QFormLayout()
        pinPropGroupBox.setLayout(fLayout)
        self.pinName = QLineEdit()
        self.pinName.setPlaceholderText("Pin Name")
        self.pinName.setToolTip("Enter pin name")
        fLayout.addRow(edf.boldLabel("Pin Name"), self.pinName)
        self.pinDir = QComboBox()
        self.pinDir.addItems(lshp.layoutPin.pinDirs)
        self.pinDir.setToolTip("Select pin direction")
        fLayout.addRow(edf.boldLabel("Pin Direction"), self.pinDir)
        self.pinType = QComboBox()
        self.pinType.addItems(lshp.layoutPin.pinTypes)
        self.pinType.setToolTip("Select pin type")
        fLayout.addRow(edf.boldLabel("Pin Type"), self.pinType)
        self.pinLayerCB = QComboBox()
        fLayout.addRow(edf.boldLabel("Pin Layer:"), self.pinLayerCB)
        self.pinBottomLeftX = edf.shortLineEdit()
        fLayout.addRow(edf.boldLabel("Pin Bottom Left X:"), self.pinBottomLeftX)
        self.pinBottomLeftY = edf.shortLineEdit()
        fLayout.addRow(edf.boldLabel("Pin Bottom Left Y:"), self.pinBottomLeftY)
        self.pinTopRightX = edf.shortLineEdit()
        fLayout.addRow(edf.boldLabel("Pin Top Right X:"), self.pinTopRightX)
        self.pinTopRightY = edf.shortLineEdit()
        fLayout.addRow(edf.boldLabel("Pin Top Right Y:"), self.pinTopRightY)
        self.mainLayout.addWidget(pinPropGroupBox)

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class createLayoutLabelDialog(QDialog):
    def __init__(self, parent) -> None:
        super().__init__(parent)
        self.setWindowTitle("Create Layout Label")
        self.setMinimumWidth(300)
        fontFamilies = QFontDatabase.families(QFontDatabase.WritingSystem.Latin)
        fixedFamilies = [
            family for family in fontFamilies if QFontDatabase.isFixedPitch(family)
        ]
        QBtn = (
                QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )

        self.mainLayout = QVBoxLayout()
        labelPropBox = QGroupBox("Label Properties")
        self.labelPropLayout = QFormLayout()
        labelPropBox.setLayout(self.labelPropLayout)
        self.labelName = QLineEdit()
        self.labelName.setPlaceholderText("Label Name")
        self.labelName.setToolTip("Enter label name")
        self.labelPropLayout.addRow(edf.boldLabel("Label Name"), self.labelName)
        self.labelLayerCB = QComboBox()
        self.labelPropLayout.addRow(edf.boldLabel("Label Layer:"), self.labelLayerCB)
        self.familyCB = QComboBox()
        self.familyCB.addItems(fixedFamilies)
        self.familyCB.currentTextChanged.connect(self.familyFontStyles)
        self.labelPropLayout.addRow(edf.boldLabel("Font Name"), self.familyCB)
        self.fontStyleCB = QComboBox()
        self.fontStyles = QFontDatabase.styles(fixedFamilies[0])
        self.fontStyleCB.addItems(self.fontStyles)
        self.fontStyleCB.currentTextChanged.connect(self.styleFontSizes)
        self.labelPropLayout.addRow(edf.boldLabel("Font Style"), self.fontStyleCB)
        self.labelHeightCB = QComboBox()
        self.fontSizes = [
            str(size)
            for size in QFontDatabase.pointSizes(fixedFamilies[0], self.fontStyles[0])
        ]
        self.labelHeightCB.addItems(self.fontSizes)
        self.labelPropLayout.addRow(edf.boldLabel("Label Height"), self.labelHeightCB)
        self.labelAlignCB = QComboBox()
        self.labelAlignCB.addItems(lshp.layoutLabel.LABEL_ALIGNMENTS)
        self.labelPropLayout.addRow(edf.boldLabel("Label Alignment"), self.labelAlignCB)
        self.labelOrientCB = QComboBox()
        self.labelOrientCB.addItems(lshp.layoutLabel.LABEL_ORIENTS)
        self.labelPropLayout.addRow(
            edf.boldLabel("Label Orientation"), self.labelOrientCB
        )
        self.mainLayout.addWidget(labelPropBox)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()

    def familyFontStyles(self, s):
        self.fontStyleCB.clear()
        self.fontStyles = QFontDatabase.styles(self.familyCB.currentText())
        self.fontStyleCB.addItems(self.fontStyles)

    def styleFontSizes(self, s):
        selectedFamily = self.familyCB.currentText()
        selectedStyle = self.fontStyleCB.currentText()
        self.fontSizes = [
            str(size)
            for size in QFontDatabase.pointSizes(selectedFamily, selectedStyle)
        ]
        self.labelHeightCB.clear()
        self.labelHeightCB.addItems(self.fontSizes)


class layoutLabelProperties(createLayoutLabelDialog):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setWindowTitle("Layout Label Properties")
        self.labelTopLeftX = edf.shortLineEdit()
        self.labelPropLayout.addRow(
            edf.boldLabel("Label Top Left X:"), self.labelTopLeftX
        )
        self.labelTopLeftY = edf.shortLineEdit()
        self.labelPropLayout.addRow(
            edf.boldLabel("Label Top Left Y:"), self.labelTopLeftY
        )


class createLayoutViaDialog(QDialog):
    def __init__(self, parent) -> None:
        super().__init__(parent)
        self._parent = parent
        self.setWindowTitle("Create Via(s)")
        self.setMinimumWidth(300)

        mainLayout = QVBoxLayout()
        self.viaTypeGroup = QGroupBox("Via Type")
        self.viaTypeLayout = QHBoxLayout()
        self.singleViaRB = QRadioButton("Single")
        self.singleViaRB.setChecked(True)
        self.singleViaRB.clicked.connect(self.singleViaClicked)
        self.arrayViaRB = QRadioButton("Array")
        self.arrayViaRB.clicked.connect(self.arrayViaClicked)
        self.viaTypeLayout.addWidget(self.singleViaRB)
        self.viaTypeLayout.addWidget(self.arrayViaRB)
        self.viaTypeGroup.setLayout(self.viaTypeLayout)
        mainLayout.addWidget(self.viaTypeGroup)
        self.singleViaPropsGroup = QGroupBox("Single Via Properties")
        singleViaPropsLayout = QFormLayout()
        self.singleViaPropsGroup.setLayout(singleViaPropsLayout)
        self.singleViaNamesCB = QComboBox()
        self.singleViaNamesCB.currentTextChanged.connect(self.singleViaNameChanged)
        singleViaPropsLayout.addRow(edf.boldLabel("Via Name"), self.singleViaNamesCB)
        self.singleViaWidthEdit = edf.shortLineEdit()

        self.singleViaWidthEdit.editingFinished.connect(self.singleViaWidthChanged)
        singleViaPropsLayout.addRow(edf.boldLabel("Via Width"), self.singleViaWidthEdit)
        self.singleViaHeightEdit = edf.shortLineEdit()

        self.singleViaHeightEdit.editingFinished.connect(self.singleViaHeightChanged)
        singleViaPropsLayout.addRow(
            edf.boldLabel("Via Height"), self.singleViaHeightEdit
        )
        mainLayout.addWidget(self.singleViaPropsGroup)
        self.arrayViaPropsGroup = QGroupBox("Single Via Properties")
        arrayViaPropsLayout = QFormLayout()
        self.arrayViaPropsGroup.setLayout(arrayViaPropsLayout)
        self.arrayViaNamesCB = QComboBox()

        self.arrayViaNamesCB.currentTextChanged.connect(self.arrayViaNameChanged)
        arrayViaPropsLayout.addRow(edf.boldLabel("Via Name"), self.arrayViaNamesCB)
        self.arrayViaWidthEdit = edf.shortLineEdit()

        self.arrayViaWidthEdit.editingFinished.connect(self.arrayViaWidthChanged)
        arrayViaPropsLayout.addRow(edf.boldLabel("Via Width"), self.arrayViaWidthEdit)
        self.arrayViaHeightEdit = edf.shortLineEdit()

        self.singleViaHeightEdit.editingFinished.connect(self.arrayViaHeightChanged)
        arrayViaPropsLayout.addRow(edf.boldLabel("Via Height"), self.arrayViaHeightEdit)
        self.arrayXspacingEdit = edf.shortLineEdit()
        self.arrayXspacingEdit.editingFinished.connect(
            lambda: self.arrayViaSpacingChanged(self.arrayXspacingEdit)
        )
        arrayViaPropsLayout.addRow(
            edf.boldLabel("Column Spacing"), self.arrayXspacingEdit
        )
        self.arrayYspacingEdit = edf.shortLineEdit()
        self.arrayYspacingEdit.editingFinished.connect(
            lambda: self.arrayViaSpacingChanged(self.arrayYspacingEdit)
        )
        arrayViaPropsLayout.addRow(edf.boldLabel("Row Spacing"), self.arrayYspacingEdit)
        self.arrayXNumEdit = edf.shortLineEdit()
        self.arrayXNumEdit.setText("1")
        arrayViaPropsLayout.addRow(
            edf.boldLabel("Number of Columns"), self.arrayXNumEdit
        )
        self.arrayYNumEdit = edf.shortLineEdit()
        self.arrayYNumEdit.setText("1")
        arrayViaPropsLayout.addRow(edf.boldLabel("Number of Rows:"), self.arrayYNumEdit)
        mainLayout.addWidget(self.arrayViaPropsGroup)
        self.arrayViaPropsGroup.hide()
        self.singleViaPropsGroup.show()

        self.viaLocationGroup = QGroupBox("Via Location")
        self.viaLocationLayout = QFormLayout()
        self.viaLocationGroup.setLayout(self.viaLocationLayout)
        self.startXEdit = edf.shortLineEdit()
        self.viaLocationLayout.addRow(edf.boldLabel("Start X:"), self.startXEdit)
        self.startYEdit = edf.shortLineEdit()
        self.viaLocationLayout.addRow(edf.boldLabel("Start Y:"), self.startYEdit)
        mainLayout.addWidget(self.viaLocationGroup)
        self.viaLocationGroup.hide()

        QBtn = (
                QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)
        self.show()

    def singleViaClicked(self):
        self.arrayViaPropsGroup.hide()
        self.singleViaPropsGroup.show()
        self.adjustSize()

    def arrayViaClicked(self):
        self.singleViaPropsGroup.hide()
        self.arrayViaPropsGroup.show()
        self.adjustSize()

    def singleViaNameChanged(self, text: str):
        via = [item for item in process.processVias if item.name == text][0]
        self.singleViaWidthEdit.setText(str(via.minWidth))
        self.singleViaHeightEdit.setText(str(via.minHeight))

    def arrayViaNameChanged(self, text: str):
        via = [item for item in process.processVias if item.name == text][0]
        self.arrayViaWidthEdit.setText(str(via.minWidth))
        self.arrayViaHeightEdit.setText(str(via.minWidth))

    def singleViaWidthChanged(self):
        text = self.singleViaWidthEdit.text()
        viaDefTuple = [
            item
            for item in process.processVias
            if item.name == self.singleViaNamesCB.currentText()
        ][0]
        self.validateValue(
            text, self.singleViaWidthEdit, viaDefTuple.minWidth, viaDefTuple.maxWidth
        )

    def singleViaHeightChanged(self):
        text = self.singleViaHeightEdit.text()
        viaDefTuple = [
            item
            for item in process.processVias
            if item.name == self.singleViaNamesCB.currentText()
        ][0]
        self.validateValue(
            text, self.singleViaHeightEdit, viaDefTuple.minHeight, viaDefTuple.maxHeight
        )

    def arrayViaWidthChanged(self):
        text = self.arrayViaWidthEdit.text()
        viaDefTuple = [
            item
            for item in process.processVias
            if item.name == self.arrayViaNamesCB.currentText()
        ][0]
        self.validateValue(
            text, self.arrayViaWidthEdit, viaDefTuple.minWidth, viaDefTuple.maxWidth
        )

    def arrayViaHeightChanged(self):
        text = self.arrayViaHeightEdit.text()
        viaDefTuple = [
            item
            for item in process.processVias
            if item.name == self.arrayViaNamesCB.currentText()
        ][0]
        self.validateValue(
            text, self.arrayViaHeightEdit, viaDefTuple.minHeight, viaDefTuple.maxHeight
        )

    def arrayViaSpacingChanged(self, spaceEditField):
        text = spaceEditField.text()
        viaDefTuple = [
            item
            for item in process.processVias
            if item.name == self.arrayViaNamesCB.currentText()
        ][0]
        self.validateValue(
            text,
            spaceEditField,
            viaDefTuple.minSpacing,
            viaDefTuple.maxSpacing,
        )

    def validateValue(
            self, text: str, lineEdit: QLineEdit, min_val: float, max_val: float
    ):
        if not text:
            lineEdit.setText(str(min_val))
            return

        try:
            value = float(text)
            if value < min_val:
                self._parent.logger.warning(f"Value too small, set back to {min_val}")
                lineEdit.setText(str(min_val))
            elif value > max_val:
                self._parent.logger.warning(f"Value too large, set back to {max_val}")
                lineEdit.setText(str(max_val))
        except ValueError:
            self._parent.logger.warning(f"Invalid number format, set back to {min_val}")
            lineEdit.setText(str(min_val))


class layoutViaProperties(createLayoutViaDialog):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setWindowTitle("Layout Via Properties")

        self.viaLocationGroup.show()
        self.show()


class layoutRectProperties(QDialog):
    def __init__(self, parent: QWidget):
        super().__init__(parent)
        self.setWindowTitle("Layout Rectangle Properties")
        self.setMinimumWidth(300)
        QBtn = (
                QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout = QVBoxLayout()
        self.rectGroup = QGroupBox("Rectangle Properties")
        self.rectGroupLayout = QFormLayout()
        self.rectGroup.setLayout(self.rectGroupLayout)
        self.rectLayerCB = QComboBox()
        self.rectGroupLayout.addRow(edf.boldLabel("Rectangle Layer:"), self.rectLayerCB)
        self.rectWidthEdit = edf.shortLineEdit()
        self.rectGroupLayout.addRow(edf.boldLabel("Width:"), self.rectWidthEdit)
        self.rectHeightEdit = edf.shortLineEdit()
        self.rectGroupLayout.addRow(edf.boldLabel("Height:"), self.rectHeightEdit)
        self.topLeftEditX = edf.shortLineEdit()
        self.rectGroupLayout.addRow(edf.boldLabel("Top Left X:"), self.topLeftEditX)
        self.topLeftEditY = edf.shortLineEdit()
        self.rectGroupLayout.addRow(edf.boldLabel("Top Left Y:"), self.topLeftEditY)
        mainLayout.addWidget(self.rectGroup)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)
        self.show()


class pointsTableWidget(QTableWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setColumnCount(3)
        self.setHorizontalHeaderLabels(["Del.", "X", "Y"])
        self.setColumnWidth(0, 8)
        self.setShowGrid(True)
        self.setGridStyle(Qt.PenStyle.SolidLine)


class layoutPolygonProperties(QDialog):
    def __init__(self, parent: QWidget, tupleList: list):
        super().__init__(parent)
        self.tupleList = tupleList
        self.setWindowTitle("Layout Polygon Properties")
        self.setMinimumWidth(300)
        self.setMinimumHeight(400)
        QBtn = (
                QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout = QVBoxLayout()
        self.setLayout(mainLayout)
        polygonLayerGroup = QGroupBox("Polygon Layer")
        polygonLayerGroupLayout = QFormLayout()
        self.polygonLayerCB = QComboBox()
        polygonLayerGroupLayout.addRow(edf.boldLabel("Layer:"), self.polygonLayerCB)
        mainLayout.addLayout(polygonLayerGroupLayout)
        self.tableWidget = pointsTableWidget(self)
        mainLayout.addWidget(self.tableWidget)
        mainLayout.addWidget(self.buttonBox)

        self.populateTable()

    def populateTable(self):
        self.tableWidget.setRowCount(len(self.tupleList) + 1)  # Add one extra row

        for row, item in enumerate(self.tupleList):
            self.addRow(row, item)

        # Add an empty row at the end
        self.addEmptyRow(len(self.tupleList))

        # Connect cellChanged signal to handle when the last row is edited
        self.tableWidget.cellChanged.connect(self.handleCellChange)

    def addRow(self, row, item):

        delete_checkbox = QCheckBox()
        self.tableWidget.setCellWidget(row, 0, delete_checkbox)

        self.tableWidget.setItem(row, 1, QTableWidgetItem(str(item[0])))
        self.tableWidget.setItem(row, 2, QTableWidgetItem(str(item[1])))
        delete_checkbox.stateChanged.connect(
            lambda state, r=row: self.deleteRow(r, state)
        )

    def addEmptyRow(self, row):

        # self.table_widget.insertRow(row)
        delete_checkbox = QCheckBox()
        self.tableWidget.setCellWidget(row, 0, delete_checkbox)
        delete_checkbox.stateChanged.connect(
            lambda state, r=row: self.deleteRow(r, state)
        )

        self.tableWidget.setItem(row, 1, QTableWidgetItem(""))
        self.tableWidget.setItem(row, 2, QTableWidgetItem(""))

    def handleCellChange(self, row, column):
        if (
                row == self.tableWidget.rowCount() - 1
        ):  # Check if last row and tuple text column
            item1 = self.tableWidget.item(row, 1)
            item2 = self.tableWidget.item(row, 2)
            if item1 is not None and item2 is not None:
                text1 = item1.text()
                text2 = item2.text()
                if text1 != "" and text2 != "":
                    self.tableWidget.insertRow(row + 1)
                    self.addEmptyRow(row + 1)

    def deleteRow(self, row, state):
        # print("delete")
        if state == 2:  # Checked state
            self.tableWidget.removeRow(row)


class formDictionary:
    """
    This code defines a utility class formDictionary that extracts data
    from a Qt form layout and converts it into a Python dictionary.
    """

    def __init__(self, formLayout: QFormLayout):
        self.formLayout = formLayout

    def extractDictFormLayout(self) -> Dict[str, edf.longLineEdit]:
        data = {}
        for row in range(self.formLayout.rowCount()):
            labelItem = self.formLayout.itemAt(row, QFormLayout.ItemRole.LabelRole)
            fieldItem = self.formLayout.itemAt(row, QFormLayout.ItemRole.FieldRole)

            if labelItem and fieldItem:
                label = labelItem.widget()
                field = fieldItem.widget()

                if isinstance(label, QLabel) and isinstance(field, QLineEdit):
                    key = label.text().rstrip(":")  # Remove trailing colon if present
                    value = field
                    data[key] = value
        return data


class drcErrorsDialogue(QDialog):

    def __init__(self, parent, drcDataPathObj: Path):
        super().__init__(parent)
        self.setWindowTitle("DRC Errors Table")
        self.setMinimumSize(1200, 600)
        self.setWindowFlags(self.windowFlags() | Qt.WindowType.Window)

        layout = QVBoxLayout()

        # Add DRC table view
        import revedaEditor.fileio.importlyrdb as imlyrdb

        DRCDataObj = imlyrdb.DRCOutput(str(drcDataPathObj))
        DRCDataObj.parseDRCOutput()
        self.drcTable = drcmv.DRCTableView(
            DRCDataObj.result["violations"], DRCDataObj.result["categories"]
        )
        layout.addWidget(self.drcTable)

        QBtn = (
                QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )

        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        layout.addWidget(self.buttonBox)

        self.setLayout(layout)

        # Connect signal for polygon highlighting
        # self.drcTable.polygonSelected.connect(self.highlightPolygons)

    # def highlightPolygons(self, polygons):
    #     # Emit signal or call parentW method to highlight polygons in scene
    #     if hasattr(self.parentW(), 'highlightDRCPolygons'):
    #         self.parentW().highlightDRCPolygons(polygons)



#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting) a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#
#   Add-ons and extensions developed for this software may be distributed
#   under their own separate licenses.
#
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

import pathlib
import time
from contextlib import contextmanager
from itertools import cycle
from typing import TYPE_CHECKING, List, Union

from PySide6.QtCore import (QEvent, QPoint, QRectF, QSizeF, Qt, QRect)
from PySide6.QtGui import QColor, QGuiApplication, QPainterPath, QPen, QTransform
from PySide6.QtWidgets import (QApplication, QCompleter, QDialog, QGraphicsItem,
                               QGraphicsRectItem, QGraphicsScene, QMenu, )

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.undoStack as us
import revedaEditor.gui.propertyDialogues as pdlg

if TYPE_CHECKING:
    import revedaEditor.common.layoutShapes as lshp
    import revedaEditor.common.shapes as shp


class editorScene(QGraphicsScene):
    # Define MOUSE_EVENTS as a class attribute to avoid recreating it on every call
    MOUSE_EVENTS = {QEvent.Type.GraphicsSceneMouseMove, QEvent.Type.GraphicsSceneMousePress,
                    QEvent.Type.GraphicsSceneMouseRelease, }

    def __init__(self, parent):
        super().__init__(parent)
        self.containerWidget = parent
        self.editorWindow = self.containerWidget.editorWindow

        self.majorGrid = self.editorWindow.majorGrid
        self.snapGrid = self.editorWindow.snapGrid
        self.snapTuple = self.editorWindow.snapTuple
        self.snapDistance = min(*self.editorWindow.snapTuple)

        # Initialize mouse-related attributes together
        self.mousePressLoc = self.mouseMoveLoc = self.mouseReleaseLoc = None

        # Use dictionary unpacking for edit modes
        self.editModes = ddef.editModes(
            **{"selectItem": True, "deleteItem": False, "moveItem": False,
               "copyItem": False, "rotateItem": False, "changeOrigin": False,
               "zoomView": False, "panView": False, "stretchItem": False,
               "alignItems": False, })

        self.messages = {"selectItem": "Select Item", "deleteItem": "Delete Item",
                         "moveItem": "Move Item", "copyItem": "Copy Item",
                         "rotateItem": "Rotate Item",
                         "changeOrigin": "Change Origin",
                         "panView": "Pan View at Mouse Press Position",
                         "zoomView": "Draw a rectangle to zoom in",
                         "stretchItem": "Stretch Item",
                         "alignItems": "Select Alignment Option.", }
        # Initialize undo stack with limit
        self.undoStack = us.undoStack()
        self.undoStack.setUndoLimit(99)
        self.itemsRefSet: set[QGraphicsItem] = set()

        # Group selection-related attributes
        self.partialSelection = False
        self.selectionRectItem: Union[QGraphicsRectItem, None] = None
        self.zoomRectItem: Union[QGraphicsRectItem, None] = None
        self.selectedItemsSet: set[QGraphicsItem] = set()
        self.selectedItemGroup = None
        self.itemCycler = None
        self._groupItems = []
        self.itemsAtPressSet = set()
        self._draftPen = QPen(QColor(0, 150, 0, 128), int(self.snapGrid / 2), Qt.DashLine)
        self._zoomPen = QPen(QColor(255, 0, 0, 255), 2, Qt.DashLine)
        self._draftPen.setCosmetic(False)
        self._zoomPen.setCosmetic(False)

        # Initialize UI elements
        self.origin = QPoint(0, 0)
        self.cellName = self.editorWindow.file.parent.stem
        self.libraryDict = self.editorWindow.libraryDict
        self.itemContextMenu = QMenu()

        # Get application-level references
        app_main = self.editorWindow.appMainW
        self.appMainW = app_main
        self.logger = app_main.logger
        self.messageLine = self.editorWindow.messageLine
        self.statusLine = self.editorWindow.statusLine

        # Scene properties
        self.readOnly = False
        self.installEventFilter(self)
        self.setMinimumRenderSize(2)
        self._initialGroupPosList = []
        self._initialGroupPos = QPoint(0, 0)
        self._finalGroupPosDiff = QPoint(0, 0)
        self.itemCycler = None
        self.mousePressLoc = QPoint(0, 0)
        self.mouseMoveLoc = QPoint(0, 0)
        self.mouseReleaseLoc = QPoint(0, 0)
        self.newAlignLine = None

    def contextMenuEvent(self, event):
        if self.itemAt(event.scenePos(), QTransform()) is None:
            self.messageLine.setText("No item selected")
        super().contextMenuEvent(event)

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        modifiers = QGuiApplication.keyboardModifiers()
        if event.button() == Qt.MouseButton.LeftButton:
            self.mousePressLoc = event.scenePos().toPoint()
            if self.editModes.selectItem:
                if (self.selectionRectItem is None and self.itemAt(self.mousePressLoc,
                                                                   QTransform()) is None):
                    self.selectionRectItem = QGraphicsRectItem(self.mousePressLoc.x(),
                                                               self.mousePressLoc.y(), 0, 0)
                    self.selectionRectItem.setPen(self._draftPen)  # Use property
                    self.selectionRectItem.setZValue(100)
                    self.addItem(self.selectionRectItem)
                elif self.itemAt(self.mousePressLoc, QTransform()):
                    self.itemsAtPressSet = {item for item in self.items(self.mousePressLoc)
                                            if item.parentItem() is None}
                    if self.itemsAtPressSet:
                        self.itemCycler = cycle(self.itemsAtPressSet)
                        item = next(self.itemCycler)
                        if modifiers == Qt.KeyboardModifier.ShiftModifier:
                            self.selectedItemsSet |= {item}
                            item.setSelected(True)
                        elif modifiers == Qt.KeyboardModifier.ControlModifier:
                            if item in self.selectedItemsSet:
                                self.selectedItemsSet.remove(item)
                                item.setSelected(False)
                            else:
                                self.selectedItemsSet.add(item)
                                item.setSelected(True)
                        elif modifiers == Qt.KeyboardModifier.NoModifier:
                            self.clearSelection()
                            self.selectedItemsSet = {item}
                            item.setSelected(True)
        

            if self.editModes.moveItem:
                self.selectedItemGroup = self.createItemGroup(self.selectedItems())
                self._initialGroupPos = self.selectedItemGroup.pos().toPoint()
                self._initialGroupPosList = []
                for item in self.selectedItemGroup.childItems():
                    # Disable individual item movement so Qt's drag-tracking (which
                    # stored press offsets before the group existed) cannot apply
                    # stale, wrong offsets that would cause the items to jump.
                    # The scene moves the whole group manually in mouseMoveEvent.
                    item.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable, False)
                    self._initialGroupPosList.append(item.scenePos().toPoint())
            elif self.editModes.panView:
                self.centerViewOnPoint(self.mousePressLoc)
            elif self.editModes.zoomView:
                if self.zoomRectItem is None:
                    self.zoomRectItem = QGraphicsRectItem(self.mousePressLoc.x(),
                                                          self.mousePressLoc.y(), 0, 0)
                    self.zoomRectItem.setPen(self._zoomPen)  # Use property
                    self.zoomRectItem.setZValue(100)
                    self.addItem(self.zoomRectItem)
            self.messageLine.setText(self.messages.get(self.editModes.mode(), ""))

    def mouseMoveEvent(self, event):
        super().mouseMoveEvent(event)
        self.mouseMoveLoc = event.scenePos().toPoint()
        if self.editModes.selectItem:
            if self.selectionRectItem:
                self.selectionRectItem.setRect(QRectF(self.mousePressLoc.toPointF(),
                                                      self.mouseMoveLoc.toPointF()).normalized())
        elif self.editModes.zoomView:
            if self.zoomRectItem:
                self.zoomRectItem.setRect(QRectF(self.mousePressLoc.toPointF(),
                                                 self.mouseMoveLoc.toPointF()).normalized())

        elif self.editModes.moveItem and self.selectedItemGroup:
            offset = self.mouseMoveLoc - self.mousePressLoc
            self.selectedItemGroup.setPos(offset)
        elif self.editModes.copyItem and self.selectedItemGroup:
            offset = self.mouseMoveLoc - self.mousePressLoc
            self.selectedItemGroup.setPos(offset)

    def mouseReleaseEvent(self, event):
        # if event.button() != Qt.MouseButton.LeftButton:
        #     super().mouseReleaseEvent(event)
        #     return

        modifiers = QGuiApplication.keyboardModifiers()
        self.mouseReleaseLoc = event.scenePos().toPoint()

        if self.editModes.moveItem and self.selectedItemGroup:
            _groupItems = self.selectedItemGroup.childItems()
            # Calculate final scene positions before destroying group
            _finalGroupPosList = [item.scenePos().toPoint() for item in _groupItems]
            _posDiff = [finalPos - initPos for finalPos, initPos in zip(_finalGroupPosList, self._initialGroupPosList)]
            
            self.destroyItemGroup(self.selectedItemGroup)
            self.selectedItemGroup = None
            
            # Use the actual position difference for each item
            if _posDiff and _posDiff[0] != QPoint(0, 0):
                self.undoGroupMoveStack(_groupItems, self._initialGroupPosList, _posDiff[0])
            
            [item.setSelected(False) for item in _groupItems]
            self.editModes.setMode("selectItem")
        elif self.editModes.copyItem and self.selectedItemGroup:
            _groupItems = self.selectedItemGroup.childItems()
            self.destroyItemGroup(self.selectedItemGroup)
            self.selectedItemGroup = None
            for item in _groupItems:
                item.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable, True)
            self.editModes.setMode("selectItem")
        elif self.editModes.selectItem:
            if self.selectionRectItem:
                selectionPath = QPainterPath()
                selectionPath.addRect(self.selectionRectItem.sceneBoundingRect())
                selectionMode = (
                    Qt.ItemSelectionMode.IntersectsItemShape if self.partialSelection else Qt.ItemSelectionMode.ContainsItemShape)
                itemsInRectSet = set(
                    [item for item in self.items(selectionPath, mode=selectionMode) if
                     item.parentItem() is None])
                if modifiers == Qt.KeyboardModifier.ShiftModifier:
                    self.selectedItemsSet |= itemsInRectSet
                elif modifiers == Qt.KeyboardModifier.ControlModifier:
                    for item in itemsInRectSet:
                        item.setSelected(False)
                    self.selectedItemsSet -= itemsInRectSet
                elif modifiers == Qt.KeyboardModifier.NoModifier:
                    self.selectedItemsSet = itemsInRectSet
                for item in self.selectedItemsSet:
                    item.setSelected(True)
                self.removeItem(self.selectionRectItem)
                self.selectionRectItem = None
            elif not self.itemsAtPressSet:
                self.itemCycler = None
                self.selectedItemsSet = set()
                self.deselectAll()
                self.clearSelection()
        elif self.editModes.zoomView and self.zoomRectItem:
            self.zoomToRect(self.zoomRectItem.rect().toRect())
            self.editModes.setMode('selectItem')
            self.removeItem(self.zoomRectItem)
            self.zoomRectItem = None
        elif self.editModes.changeOrigin:
            self.origin = self.mouseReleaseLoc
            self.editModes.setMode('selectItem')
        else:
            self._handleMouseRelease(self.mouseReleaseLoc, event.button())
        self.messageLine.setText(self.messages.get(self.editModes.mode(), ""))
        super().mouseReleaseEvent(event)

    def snapToGrid(self, point: QPoint) -> QPoint:
        """Snap point to scene grid."""
        xgrid = self.snapTuple[0]
        ygrid = self.snapTuple[1]
        return QPoint(round(point.x() / xgrid) * xgrid, round(point.y() / ygrid) * ygrid)

    def _snapPoint(self, pos: QPoint) -> QPoint:
        """
        Default snapping behavior. Subclasses can override this for more
        advanced snapping (e.g., to items, intersections).
        """
        return self.snapToGrid(pos)

    def _handleMouseRelease(self, mousePos: QPoint, button: Qt.MouseButton):
        pass

    def snapToBase(self, number, base):
        """
        Restrict a number to the multiples of base
        """
        return int(round(float(number) / base)) * base

    def rotateSelectedItems(self, point: QPoint):
        """
        Rotate selected items by 90 degree.
        """
        for item in self.selectedItems():
            self.rotateAnItem(point, item, 90)
        self.editModes.setMode("selectItem")

    def rotateAnItem(self, point: QPoint,
                     item: Union["shp.symbolShape", "lshp.layoutShape"], angle: int, ):
        """
        Rotate a graphics item around a point by a specified angle with undo support.

        Args:
            point (QPoint): The pivot point for rotation
            item (QGraphicsItem): The item to be rotated
            angle (int): The rotation angle in degrees

        Returns:
            None
        """
        undoCommand = us.undoRotateShape(self, item, point, angle)
        self.undoStack.push(undoCommand)

    def eventFilter(self, source, event):
        """
        Filter mouse events to snap them to background grid points.
        """
        if self.readOnly:
            return True

        if event.type() in self.MOUSE_EVENTS:
            # Use the _snapPoint method which can be overridden by subclasses
            snappedPos = self._snapPoint(event.scenePos().toPoint())
            event.setScenePos(snappedPos)
            return False

        return super().eventFilter(source, event)

    def copySelectedItems(self):
        """
        Will be implemented in the subclasses.
        """

    def flipHorizontal(self):
        for item in self.selectedItems():
            item.flipTuple = (-1, 1)

    def flipVertical(self):
        for item in self.selectedItems():
            item.flipTuple = (1, -1)

    def selectAll(self):
        """
        Select all items in the scene.
        """
        [item.setSelected(True) for item in self.items()]

    def deselectAll(self):
        """
        Deselect all items in the scene.
        """
        for item in self.selectedItems():
            item.setSelected(False)
        self.clearSelection()
        self.selectedItemsSet.clear()
        self.itemsAtPressSet.clear()
        self.selectionRectItem = None
        self.zoomRectItem = None
        self.selectedItemGroup = None
        self.itemCycler = None
        self.newAlignLine = None
        self.newNet = None
        # self.stretchNet = None
        self.newInstance = None
        self.newPin = None
        self.newText = None
        self.editModes.setMode("selectItem")
        self.messageLine.setText("All items deselected")

    def deleteSelectedItems(self):
        if self.selectedItems() is not None:
            for item in self.selectedItems():
                # self.removeItem(item)
                undoCommand = us.deleteShapeUndo(self, item)
                self.undoStack.push(undoCommand)
            self.update()  # update the scene

    def stretchSelectedItems(self):
        for item in self.selectedItems():
            if hasattr(item, "stretch"):
                item.stretch = True

    def reloadScene(self):
        """Reload scene with proper painter state management."""
        self._safeLoadDesign(self.editorWindow.file, reload=True)

    def _safeLoadDesign(self, file, reload=False):
        """Safely load design with painter error prevention."""
        # Disable all updates and painting
        for view in self.views():
            view.setUpdatesEnabled(False)
            view.viewport().setUpdatesEnabled(False)

        self.blockSignals(True)

        try:
            if reload:
                self.clear()

            # Call subclass-specific loadDesign
            self.loadDesign(file)

            # Update scene rect after loading
            if self.items():
                self.setSceneRect(self.itemsBoundingRect())

        finally:
            self.blockSignals(False)
            for view in self.views():
                view.setUpdatesEnabled(True)
                view.viewport().setUpdatesEnabled(True)

            # Single deferred update
            QApplication.processEvents()

    def loadDesign(self, filePathObj: pathlib.Path):
        """
        Load the design from the specified file.
        Must be implemented in subclasses.
        """
        raise NotImplementedError("Subclasses must implement loadDesign")

    def fitItemsInView(self) -> None:
        self.setSceneRect(self.itemsBoundingRect().adjusted(-40, -40, 40, 40))
        self.views()[0].fitInView(self.sceneRect(), Qt.AspectRatioMode.KeepAspectRatio)
        self.views()[0].viewport().update()

    def zoomToRect(self) -> None:
        if self.zoomRectItem:
            self.setSceneRect(self.zoomRectItem.rect().adjusted(-40, -40, 40, 40))
            self.views()[0].fitInView(self.sceneRect(), Qt.AspectRatioMode.KeepAspectRatio)
            self.views()[0].viewport().update()

    def zoomInBy2(self):
        sceneRect = self.sceneRect()
        # New size is half of current (zoom by 2)
        newWidth = sceneRect.width() * 0.5
        newHeight = sceneRect.height() * 0.5

        # Prefer the last known mouse scene position; fall back to scene center
        mousePos = (self.mouseMoveLoc if self.mouseMoveLoc is not None and (
                self.mouseMoveLoc.x() != 0 or self.mouseMoveLoc.y() != 0) else sceneRect.center())

        # Center the new scene rect at the mouse position
        newSceneRect = QRectF(mousePos.x() - newWidth / 2.0, mousePos.y() - newHeight / 2.0,
                              newWidth, newHeight, )

        self.setSceneRect(newSceneRect)
        view = self.views()[0]
        view.fitInView(newSceneRect, Qt.AspectRatioMode.KeepAspectRatio)
        view.viewport().update()

    def zoomOutBy2(self):
        sceneRect = self.sceneRect()
        # New size is half of current (zoom by 2)
        newWidth = sceneRect.width() * 2
        newHeight = sceneRect.height() * 2

        # Prefer the last known mouse scene position; fall back to scene center
        mousePos = (self.mouseMoveLoc if self.mouseMoveLoc is not None and (
                self.mouseMoveLoc.x() != 0 or self.mouseMoveLoc.y() != 0) else sceneRect.center())

        # Center the new scene rect at the mouse position
        newSceneRect = QRectF(mousePos.x() - newWidth / 2.0, mousePos.y() - newHeight / 2.0,
                              newWidth, newHeight, )

        self.setSceneRect(newSceneRect)
        view = self.views()[0]
        view.fitInView(newSceneRect, Qt.AspectRatioMode.KeepAspectRatio)
        view.viewport().update()

    def zoomToRect(self, newSceneRect:QRect):
        self.setSceneRect(newSceneRect)
        view = self.views()[0]
        view.fitInView(newSceneRect, Qt.AspectRatioMode.KeepAspectRatio)
        view.viewport().update()    

    def moveSceneLeft(self) -> None:
        currentSceneRect = self.sceneRect()
        halfWidth = currentSceneRect.width() / 2.0
        newSceneRect = QRectF(currentSceneRect.left() - halfWidth, currentSceneRect.top(),
                              currentSceneRect.width(), currentSceneRect.height(), )
        self.setSceneRect(newSceneRect)

    def moveSceneRight(self) -> None:
        currentSceneRect = self.sceneRect()
        halfWidth = currentSceneRect.width() / 2.0
        newSceneRect = QRectF(currentSceneRect.left() + halfWidth, currentSceneRect.top(),
                              currentSceneRect.width(), currentSceneRect.height(), )
        self.setSceneRect(newSceneRect)

    def moveSceneUp(self) -> None:
        currentSceneRect = self.sceneRect()
        halfWidth = currentSceneRect.width() / 2.0
        newSceneRect = QRectF(currentSceneRect.left(), currentSceneRect.top() - halfWidth,
                              currentSceneRect.width(), currentSceneRect.height(), )
        self.setSceneRect(newSceneRect)

    def moveSceneDown(self) -> None:
        currentSceneRect = self.sceneRect()
        halfWidth = currentSceneRect.width() / 2.0
        newSceneRect = QRectF(currentSceneRect.left(), currentSceneRect.top() + halfWidth,
                              currentSceneRect.width(), currentSceneRect.height(), )
        self.setSceneRect(newSceneRect)

    def centerViewOnPoint(self, point: QPoint) -> None:
        currentSceneRect = self.sceneRect()
        size = QSizeF(currentSceneRect.width(), currentSceneRect.height())
        newSceneRect = QRectF(point.x() - size.width() / 2, point.y() - size.height() / 2,
                              size.width(), size.height(), )
        self.setSceneRect(newSceneRect)

    def addUndoStack(self, item: QGraphicsItem):
        undoCommand = us.addShapeUndo(self, item)
        self.undoStack.push(undoCommand)

    def deleteUndoStack(self, item: QGraphicsItem):
        undoCommand = us.deleteShapeUndo(self, item)
        self.undoStack.push(undoCommand)

    def addListUndoStack(self, itemList: List[QGraphicsItem]) -> None:
        undoCommand = us.addShapesUndo(self, itemList)
        self.undoStack.push(undoCommand)

    def deleteListUndoStack(self, itemList: List[QGraphicsItem]) -> None:
        undoCommand = us.deleteShapesUndo(self, itemList)
        self.undoStack.push(undoCommand)

    def undoGroupMoveStack(self, items: List[QGraphicsItem], startPos: List[QPoint],
                           endPos: QPoint) -> None:
        undoCommand = us.undoGroupMove(self, items, startPos, endPos)
        self.undoStack.push(undoCommand)

    def addUndoMacroStack(self, undoCommands: list, macroName: str = "Macro"):
        self.undoStack.beginMacro(macroName)
        for command in undoCommands:
            self.undoStack.push(command)
        self.undoStack.endMacro()

    def moveBySelectedItems(self):
        if self.selectedItems():
            dlg = pdlg.moveByDialogue(self.editorWindow)
            dlg.xEdit.setText("0")
            dlg.yEdit.setText("0")
            if dlg.exec() == QDialog.DialogCode.Accepted:
                dx = self.snapToBase(float(dlg.xEdit.text()), self.snapTuple[0])
                dy = self.snapToBase(float(dlg.yEdit.text()), self.snapTuple[1])
                moveCommand = us.undoMoveByCommand(self, self.selectedItems(), dx, dy)
                self.undoStack.push(moveCommand)
                self.editorWindow.messageLine.setText(
                    f"Moved items by {dlg.xEdit.text()} and {dlg.yEdit.text()}")
                self.editModes.setMode("selectItem")

    def cellNameComplete(self, dlg: QDialog, cellNameList: List[str]):
        cellNameCompleter = QCompleter(cellNameList)
        cellNameCompleter.setCaseSensitivity(Qt.CaseInsensitive)
        dlg.instanceCellName.setCompleter(cellNameCompleter)

    def viewNameComplete(self, dlg: QDialog, viewNameList: List[str]):
        viewNameCompleter = QCompleter(viewNameList)
        viewNameCompleter.setCaseSensitivity(Qt.CaseInsensitive)
        dlg.instanceViewName.setCompleter(viewNameCompleter)
        dlg.instanceViewName.setText(viewNameList[0])

    @contextmanager
    def measureDuration(self):
        start_time = time.perf_counter()
        try:
            yield
        finally:
            end_time = time.perf_counter()
            self.logger.info(f"Total processing time: {end_time - start_time:.3f} seconds")

    @property
    def draftPen(self):
        return self._draftPen
